import logging
import multiprocessing
from typing import Dict, Union

import numpy as np

from freemocap.core_processes.post_process_skeleton_data.calculate_center_of_mass import (
    calculate_center_of_mass_from_skeleton,
)
from freemocap.core_processes.post_process_skeleton_data.create_skeleton import create_skeleton_model
from freemocap.core_processes.post_process_skeleton_data.enforce_rigid_bones import enforce_rigid_bones_from_skeleton
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.system.logging.configure_logging import log_view_logging_format_string
from freemocap.system.logging.queue_logger import DirectQueueHandler

logger = logging.getLogger(__name__)


def calculate_anatomical_data(
        processing_parameters: ProcessingParameterModel,
        skel3d_frame_marker_xyz: np.ndarray,
        queue: multiprocessing.Queue,
) -> Dict[str, Union[np.ndarray, None]]:
    if queue:
        handler = DirectQueueHandler(queue)
        handler.setFormatter(logging.Formatter(fmt=log_view_logging_format_string, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(handler)

    logger.info("Creating skeleton model...")
    skeleton = create_skeleton_model(
        actual_markers=processing_parameters.tracking_model_info.landmark_names,
        segment_connections=processing_parameters.tracking_model_info.segment_connections,
        virtual_markers=processing_parameters.tracking_model_info.virtual_markers_definitions,
        joint_hierarchy=processing_parameters.tracking_model_info.joint_hierarchy,
        center_of_mass_info=processing_parameters.tracking_model_info.center_of_mass_definitions,
        num_tracked_points=processing_parameters.tracking_model_info.num_tracked_points,
    )

    skeleton.integrate_freemocap_3d_data(skel3d_frame_marker_xyz)

    logger.info("Calculating center of mass...")
    try:
        segment_COM_frame_imgPoint_XYZ, totalBodyCOM_frame_XYZ = calculate_center_of_mass_from_skeleton(
            skeleton=skeleton
        )
    except ValueError:
        logger.warning("Center of mass cannot be calculated for this tracking model")
        segment_COM_frame_imgPoint_XYZ = totalBodyCOM_frame_XYZ = None
    except AttributeError as e:
        raise e

    logger.info("Enforcing rigid bones...")
    rigid_bones_data = enforce_rigid_bones_from_skeleton(skeleton=skeleton)

    return {
        "segment_COM": segment_COM_frame_imgPoint_XYZ,
        "total_body_COM": totalBodyCOM_frame_XYZ,
        "rigid_bones_data": rigid_bones_data,
    }



import logging
import threading
from pathlib import Path
from typing import Union

from PySide6.QtCore import Signal, QThread

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)
from freemocap.core_processes.capture_volume_calibration.run_anipose_capture_volume_calibration import (
    run_anipose_capture_volume_calibration,
)

logger = logging.getLogger(__name__)


class AniposeCalibrationThreadWorker(QThread):
    finished = Signal(str)
    in_progress = Signal(str)
    groundplane_failed = Signal(str)

    def __init__(
            self,
            calibration_videos_folder_path: Union[str, Path],
            charuco_square_size: Union[int, float],
            kill_thread_event: threading.Event,
            charuco_board_definition: CharucoBoardDefinition,
            use_charuco_as_groundplane: bool = False,
    ):
        super().__init__()
        logger.info(
            f"Initializing Anipose Calibration Thread Worker for videos in path {calibration_videos_folder_path}"
        )
        # if charuco_board_definition is None: #removing because with the new dropdown, we should always have a charuco board definition
        #     charuco_board_definition = CharucoBoardDefinition()

        self._kill_thread_event = kill_thread_event
        self._charuco_board_definition = charuco_board_definition
        self._charuco_square_size = charuco_square_size
        self._calibration_videos_folder_path = calibration_videos_folder_path
        self._use_charuco_as_groundplane = use_charuco_as_groundplane
        self._work_done = False

    @property
    def work_done(self):
        return self._work_done

    def _emit_in_progress_data(self, message: str):
        self.in_progress.emit(message)

    def run(self):
        logger.info("Beginning Anipose calibration with Charuco Square Size (mm): {}".format(self._charuco_square_size))

        try:
            toml_path, groundplane_success = run_anipose_capture_volume_calibration(
                charuco_board_definition=self._charuco_board_definition,
                charuco_square_size=self._charuco_square_size,
                calibration_videos_folder_path=self._calibration_videos_folder_path,
                pin_camera_0_to_origin=True,
                use_charuco_as_groundplane=self._use_charuco_as_groundplane,
                progress_callback=self._emit_in_progress_data,
            )
            self.finished.emit(str(toml_path))
            if groundplane_success:
                if groundplane_success.success is False:
                    self.groundplane_failed.emit(
                        f"Groundplane calibration failed: {groundplane_success.error} \n\n Using original camera calibration instead."
                    )

        except Exception as e:
            logger.exception("something went wrong in the anipose calibration")
            logger.exception(e)

        self._work_done = True

        logger.info("Anipose Calibration Complete")



import colorsys

import numpy as np


def bright_color_generator():
    hue = 0  # initialize hue

    while True:
        # create a bright color by using the full saturation and value
        # hue is varied over time to generate different colors
        r, g, b = [int(255 * i) for i in colorsys.hsv_to_rgb(hue, 1, 1)]

        yield r, g, b

        # increment hue
        hue += 0.1
        # avoid red and blue areas by wrapping back to 0.2 when it reaches 0.6
        if hue >= 0.6:  # hue wraps back to 0.2 when it reaches 0.6
            hue = 0.2


bright_colors = bright_color_generator()


def get_next_color():
    return next(bright_colors)


def rgb_color_generator(start_color, end_color, phase_increment=0.01):
    r_start, g_start, b_start = start_color
    r_end, g_end, b_end = end_color

    # Calculate the range of each color
    r_range = r_end - r_start
    g_range = g_end - g_start
    b_range = b_end - b_start

    # Initialize the current phase of each color wave
    r_phase = 0
    g_phase = np.pi / 3  # Offset the green phase by 1/3 of a wave
    b_phase = 2 * np.pi / 3  # Offset the blue phase by 2/3 of a wave

    while True:
        # Calculate the current color using a sinusoidal wave
        r = int(r_start + r_range * (np.sin(r_phase) + 1) / 2)
        g = int(g_start + g_range * (np.sin(g_phase) + 1) / 2)
        b = int(b_start + b_range * (np.sin(b_phase) + 1) / 2)

        yield r, g, b

        # Advance the phase of each color wave
        r_phase += phase_increment
        g_phase += phase_increment
        b_phase += phase_increment

        # Wrap the phase back to the start if it exceeds 2*pi
        if r_phase > 2 * np.pi:
            r_phase -= 2 * np.pi
        if g_phase > 2 * np.pi:
            g_phase -= 2 * np.pi
        if b_phase > 2 * np.pi:
            b_phase -= 2 * np.pi



import logging
import shutil
import threading
from pathlib import Path
from typing import Union

from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QDoubleValidator
from PySide6.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QListWidget,
    QLabel,
    QFormLayout,
    QLineEdit,
    QTreeView,
    QPushButton,
    QDialog,
    QCheckBox,
    QButtonGroup,
    QRadioButton,
    QWidget,
    QFileSystemModel,
)
from skelly_synchronize import create_audio_debug_plots, create_brightness_debug_plots

from freemocap.gui.qt.workers.synchronize_videos_thread_worker import SynchronizeVideosThreadWorker
from freemocap.system.open_file import open_file
from freemocap.system.paths_and_filenames.file_and_folder_names import SYNCHRONIZED_VIDEOS_FOLDER_NAME
from freemocap.system.paths_and_filenames.path_getters import get_recording_session_folder_path
from freemocap.utilities.get_video_paths import get_video_paths

no_files_found_string = "No '.mp4' video files found! \n \n Note - We only look for `.mp4` files (for now). If your videos are a different format, convert them to `mp4` via online tools like `www.cloudconvert.com`, or softwares like `HandBrake`, `ffmpeg` or any video editing software"

logger = logging.getLogger(__name__)


class ImportVideosWizard(QDialog):
    folder_to_save_videos_to_selected = Signal(list, str, bool)

    def __init__(self, import_videos_path: Union[str, Path], kill_thread_event: threading.Event, parent=None):
        super().__init__(parent=parent)
        self.kill_thread_event = kill_thread_event

        self.setWindowTitle("Import Videos")

        self.import_videos_path = import_videos_path

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._import_directory_view = self._create_import_directory_view(import_videos_path)
        self._layout.addWidget(self._import_directory_view)

        self._video_file_paths = [str(path) for path in get_video_paths(path_to_video_folder=import_videos_path)]

        self._video_file_list_view = self._create_video_file_list_widget()
        self._layout.addWidget(self._video_file_list_view)

        self._form_layout = QFormLayout()
        self._layout.addLayout(self._form_layout)

        self._synchronize_videos_checkbox = QCheckBox()
        self._synchronize_videos_checkbox.toggled.connect(self._handle_synchronize_checkbox_toggled)

        self.synchronization_extension = self._create_synchronization_extension()

        self._form_layout.addRow("Synchronize videos:", self._synchronize_videos_checkbox)
        self._form_layout.addRow(self.synchronization_extension)

        self._folder_name = f"import_{Path(import_videos_path).name}"
        self._folder_name_line_edit = QLineEdit(parent=self)
        self._folder_name_line_edit.textChanged.connect(self._handle_folder_name_line_edit_changed)

        self._folder_name_line_edit.setPlaceholderText(self._folder_name)
        self._form_layout.addRow("Recording Name:", self._folder_name_line_edit)

        self._folder_where_videos_will_be_saved_to_label = QLabel(self._get_folder_videos_will_be_saved_to())
        self._form_layout.addRow("Videos will be saved to:", self._folder_where_videos_will_be_saved_to_label)

        self._continue_button = QPushButton("Continue")
        self._continue_button.isDefault()
        self._continue_button.clicked.connect(self._handle_continue_button_clicked)
        self._form_layout.addRow(self._continue_button)

    def _get_folder_videos_will_be_saved_to(self):
        return str(Path(get_recording_session_folder_path()) / self._folder_name / SYNCHRONIZED_VIDEOS_FOLDER_NAME)

    def _create_import_directory_view(self, import_videos_path: Union[str, Path]):
        self._file_system_model = QFileSystemModel()
        self._file_system_model.setRootPath(str(import_videos_path))

        self._tree_view_widget = QTreeView()
        self._layout.addWidget(self._tree_view_widget)
        self._tree_view_widget.setModel(self._file_system_model)
        self._tree_view_widget.setRootIndex(self._file_system_model.index(str(import_videos_path)))
        self._tree_view_widget.setColumnWidth(0, 250)
        self._tree_view_widget.setHeaderHidden(False)
        self._tree_view_widget.setAlternatingRowColors(True)
        self._tree_view_widget.setWindowTitle(str(import_videos_path))

        self._tree_view_widget.doubleClicked.connect(self._open_file)

        return self._tree_view_widget

    def _open_file(self):
        index = self._tree_view_widget.currentIndex()
        file_path = self._file_system_model.filePath(index)
        logger.info(f"Opening file from file_system_view_widget: {file_path}")
        open_file(file_path)

    def _create_video_file_list_widget(self):
        list_view = QListWidget()
        list_view.setWordWrap(True)
        if len(self._video_file_paths) == 0:
            list_view.addItem(no_files_found_string)
        else:
            list_view.addItems(self._video_file_paths)
        return list_view

    def _create_synchronization_extension(self):
        synchronization_extension = QWidget()

        self._synchronization_method_buttons = QButtonGroup(self)
        self._cross_correlation_radio_button = QRadioButton("Audio Cross Correlation")
        self._cross_correlation_radio_button.setChecked(True)
        self.synchronization_method = "audio"
        self._brightness_contrast_radio_button = QRadioButton("Brightness Contrast Detection")

        self._cross_correlation_radio_button.toggled.connect(self._handle_cross_correlation_radio_button_toggled)
        self._brightness_contrast_radio_button.toggled.connect(self._handle_brightness_contrast_radio_button_toggled)
        self._synchronization_method_buttons.addButton(self._cross_correlation_radio_button)
        self._synchronization_method_buttons.addButton(self._brightness_contrast_radio_button)

        extension_layout = QVBoxLayout()
        synch_button_layout = QHBoxLayout()
        synch_button_layout.addWidget(QLabel("Choose synchronization method:"))
        synch_button_layout.addWidget(self._cross_correlation_radio_button)
        synch_button_layout.addWidget(self._brightness_contrast_radio_button)

        synchronization_message = QLabel(
            " - Videos must have exactly the same video frame rates to be synchronized.\n\n - For audio cross correlation, audio tracks must have the same sample rate.\n"
        )
        synchronization_message.setWordWrap(True)
        ffmpeg_warning = QLabel(
            "<b>Warning: FFmpeg is required for synchronization, but was not found. <br><br> Please make sure</b> <a href='https://www.clipcat.com/blog/how-to-install-ffmpeg-on-ubuntu-windows-and-macos-a-complete-setup-guide/'>ffmpeg is installed and added to your path.</a><br>"
        )
        ffmpeg_warning.setWordWrap(True)
        ffmpeg_warning.setOpenExternalLinks(True)

        line_edit_layout = QHBoxLayout()
        self.brightness_contrast_threshold_line_edit = QLineEdit("Brightness Contrast Threshold:")
        self.brightness_contrast_threshold_line_edit.setText("1000")
        self.brightness_contrast_threshold_line_edit.setEnabled(False)
        self.brightness_contrast_threshold_line_edit.setFixedWidth(80)

        brightness_validator = QDoubleValidator()
        brightness_validator.setBottom(1)
        self.brightness_contrast_threshold_line_edit.setValidator(brightness_validator)

        line_edit_layout.addWidget(QLabel("Brightness Contrast Threshold:"), alignment=Qt.AlignmentFlag.AlignLeft)
        line_edit_layout.addWidget(self.brightness_contrast_threshold_line_edit, alignment=Qt.AlignmentFlag.AlignLeft)
        line_edit_layout.addWidget(QLabel("   Only applies to brightness contrast detection"),
                                   alignment=Qt.AlignmentFlag.AlignLeft)
        line_edit_layout.addStretch()

        extension_layout.addLayout(synch_button_layout)
        if shutil.which("ffmpeg") is None:
            extension_layout.addWidget(ffmpeg_warning)
            logger.warning("FFmpeg not found, synchronization will not run")
        extension_layout.addWidget(synchronization_message)
        extension_layout.addLayout(line_edit_layout)

        synchronization_extension.setLayout(extension_layout)

        synchronization_extension.hide()

        return synchronization_extension

    def _handle_folder_name_line_edit_changed(self, event):
        self._folder_name = self._folder_name_line_edit.text()
        self._folder_where_videos_will_be_saved_to_label.setText(self._get_folder_videos_will_be_saved_to())

    def _handle_continue_button_clicked(self, event):
        ffmpeg_found = shutil.which("ffmpeg") is not None
        if self._synchronize_videos_checkbox.isChecked() and ffmpeg_found:
            self.synchronize_videos_thread_worker = SynchronizeVideosThreadWorker(
                raw_video_folder_path=Path(self.import_videos_path),
                synchronized_video_folder_path=Path(self._get_folder_videos_will_be_saved_to()),
                kill_thread_event=self.kill_thread_event,
                synchronization_method=self.synchronization_method,
                brightness_contrast_threshold=float(self.brightness_contrast_threshold_line_edit.text()),
            )
            self.synchronize_videos_thread_worker.start()
            self.synchronize_videos_thread_worker.finished.connect(self._handle_video_synchronization_finished)
        elif self._synchronize_videos_checkbox.isChecked() and not ffmpeg_found:
            logger.error(
                "FFmpeg not found, cannot run synchronization or import videos - please install ffmpeg and add it to your path"
            )
        else:
            self.folder_to_save_videos_to_selected.emit(
                self._video_file_paths, self._get_folder_videos_will_be_saved_to(), False
            )
        self.accept()

    def _handle_synchronize_checkbox_toggled(self, event):
        if self._synchronize_videos_checkbox.isChecked():
            logger.info("Synchronize videos selected, videos will be synchronized before importing")
            self.synchronization_extension.setVisible(True)
        else:
            logger.info("Synchronize videos deselected, videos will not be synchronized")
            self.synchronization_extension.setVisible(False)

    def _handle_cross_correlation_radio_button_toggled(self, event):
        if self._cross_correlation_radio_button.isChecked():
            self.synchronization_method = "audio"

    def _handle_brightness_contrast_radio_button_toggled(self, event):
        if self._brightness_contrast_radio_button.isChecked():
            self.synchronization_method = "brightness"
            self.brightness_contrast_threshold_line_edit.setEnabled(True)
        else:
            self.brightness_contrast_threshold_line_edit.setEnabled(False)

    def _handle_video_synchronization_finished(self):
        self._video_file_paths = [
            str(path)
            for path in get_video_paths(path_to_video_folder=self.synchronize_videos_thread_worker.output_folder_path)
        ]
        if self.synchronization_method == "audio":
            create_audio_debug_plots(
                synchronized_video_folder_path=self.synchronize_videos_thread_worker.output_folder_path
            )
        if self.synchronization_method == "brightness":
            create_brightness_debug_plots(
                raw_video_folder_path=self.import_videos_path,
                synchronized_video_folder_path=self.synchronize_videos_thread_worker.output_folder_path,
            )
        self.folder_to_save_videos_to_selected.emit(
            self._video_file_paths, self._get_folder_videos_will_be_saved_to(), True
        )


if __name__ == "__main__":
    # from PySide6.QtWidgets import QApplication
    # import sys
    #
    # app = QApplication(sys.argv)
    import_videos_window = ImportVideosWizard(import_videos_path=Path().home())
    import_videos_window.exec()
    # import_videos_window.show()
    # sys.exit(app.exec())



import logging
from pathlib import Path
from typing import Union

logger = logging.getLogger(__name__)


def remove_empty_directories(root_dir: Union[str, Path]):
    """
    Recursively remove empty directories from the root directory
    :param root_dir: The root directory to start removing empty directories from
    """
    # logger.debug(f"Searching for empty directories in: {root_dir}")
    for path in Path(root_dir).rglob("*"):
        if path.is_dir() and not any(path.iterdir()):
            logger.info(f"Removing empty directory: {path}")
            path.rmdir()
        elif path.is_dir() and any(path.iterdir()):
            remove_empty_directories(path)
        else:
            continue



import logging
import multiprocessing
import shutil
import threading
from pathlib import Path
from typing import Union, List, Callable

from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QApplication,
    QDockWidget,
    QMainWindow,
    QFileDialog,
    QWidget,
    QHBoxLayout,
)
from skelly_viewer import SkellyViewer
from skellycam import (
    SkellyCamParameterTreeWidget,
    SkellyCamWidget,
)
from tqdm import tqdm

from freemocap.data_layer.generate_jupyter_notebook.generate_jupyter_notebook import (
    generate_jupyter_notebook,
)
from freemocap.data_layer.recording_models.post_processing_parameter_models import (
    ProcessingParameterModel,
)
from freemocap.data_layer.recording_models.recording_info_model import (
    RecordingInfoModel,
)
from freemocap.gui.qt.actions_and_menus.actions import Actions
from freemocap.gui.qt.actions_and_menus.menu_bar import MenuBar
from freemocap.gui.qt.style_sheet.css_file_watcher import CSSFileWatcher
from freemocap.gui.qt.style_sheet.scss_file_watcher import SCSSFileWatcher
from freemocap.gui.qt.style_sheet.set_css_style_sheet import apply_css_style_sheet
from freemocap.gui.qt.utilities.copy_timestamps_folder import copy_directory_if_contains_timestamps
from freemocap.gui.qt.utilities.get_qt_app import get_qt_app
from freemocap.gui.qt.utilities.save_and_load_gui_state import (
    GuiState,
    load_gui_state,
    save_gui_state,
)
from freemocap.gui.qt.utilities.update_most_recent_recording_toml import (
    update_most_recent_recording_toml,
)
from freemocap.gui.qt.widgets.active_recording_widget import ActiveRecordingInfoWidget
from freemocap.gui.qt.widgets.camera_controller_group_box import CameraControllerGroupBox
from freemocap.gui.qt.widgets.central_tab_widget import CentralTabWidget
from freemocap.gui.qt.widgets.control_panel.control_panel_dock_widget import (
    ControlPanelWidget,
)
from freemocap.gui.qt.widgets.control_panel.export_data_control_panel import VisualizationControlPanel
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.process_motion_capture_data_panel import (
    ProcessMotionCaptureDataPanel,
)
from freemocap.gui.qt.widgets.directory_view_widget import DirectoryViewWidget
from freemocap.gui.qt.widgets.home_widget import (
    HomeWidget,
)
from freemocap.gui.qt.widgets.import_videos_wizard import ImportVideosWizard
from freemocap.gui.qt.widgets.log_view_widget import LogViewWidget
from freemocap.gui.qt.widgets.opencv_conflict_dialog import OpencvConflictDialog
from freemocap.gui.qt.widgets.release_notes_dialogs.tabbed_release_notes_dialog import TabbedReleaseNotesDialog
from freemocap.gui.qt.widgets.set_data_folder_dialog import SetDataFolderDialog
from freemocap.gui.qt.widgets.welcome_screen_dialog import WelcomeScreenDialog
from freemocap.gui.qt.workers.download_sample_data_thread_worker import DownloadDataThreadWorker
from freemocap.gui.qt.workers.export_to_blender_thread_worker import ExportToBlenderThreadWorker
# reboot GUI method based on this - https://stackoverflow.com/a/56563926/14662833
from freemocap.system.open_file import open_file
from freemocap.system.paths_and_filenames.file_and_folder_names import (
    PATH_TO_FREEMOCAP_LOGO_SVG,
)
from freemocap.system.paths_and_filenames.path_getters import (
    get_recording_session_folder_path,
    get_css_stylesheet_path,
    get_scss_stylesheet_path,
    get_blender_file_path,
    get_most_recent_recording_path,
    get_gui_state_json_path,
)
from freemocap.system.user_data.pipedream_pings import PipedreamPings
from freemocap.utilities.remove_empty_directories import remove_empty_directories

EXIT_CODE_REBOOT = -123456789

logger = logging.getLogger(__name__)


class MainWindow(QMainWindow):
    def __init__(
            self,
            freemocap_data_folder_path: Union[str, Path],
            pipedream_pings: PipedreamPings,
            parent=None,
    ):
        super().__init__(parent=parent)
        self._log_view_widget = LogViewWidget(parent=self)  # start this first so it will grab the setup logs
        logger.info("Initializing FreeMoCap MainWindow")

        self.setMinimumSize(1280, 720)
        self.setWindowIcon(QIcon(PATH_TO_FREEMOCAP_LOGO_SVG))
        self.setWindowTitle("freemocap \U0001f480 \U00002728")

        dummy_widget = QWidget()
        self._layout = QHBoxLayout()
        dummy_widget.setLayout(self._layout)
        self.setCentralWidget(dummy_widget)

        self._css_file_watcher = self._set_up_stylesheet()

        self._freemocap_data_folder_path = freemocap_data_folder_path
        self._pipedream_pings = pipedream_pings

        self._gui_state: GuiState = load_gui_state(get_gui_state_json_path())

        self._kill_thread_event = multiprocessing.Event()

        self._active_recording_info_widget = ActiveRecordingInfoWidget(parent=self)
        self._active_recording_info_widget.new_active_recording_selected_signal.connect(
            self._handle_new_active_recording_selected
        )
        self._directory_view_widget = self._create_directory_view_widget()
        self._directory_view_widget.expand_directory_to_path(get_recording_session_folder_path())
        self._directory_view_widget.new_active_recording_selected_signal.connect(
            self._active_recording_info_widget.set_active_recording
        )

        self._actions = Actions(freemocap_main_window=self)

        self._menu_bar = MenuBar(actions=self._actions, parent=self)
        self.setMenuBar(self._menu_bar)

        self.statusBar().showMessage(
            "Watch the terminal output for status updates, we're working on integrating better status updates into the GUI"
        )

        self._central_tab_widget = self._create_central_tab_widget()
        self.setCentralWidget(self._central_tab_widget)

        self._tools_dock_widget = self._create_tools_dock_widget()
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self._tools_dock_widget)

        self._control_panel_widget = self._create_control_panel_widget(log_update=self._log_view_widget.add_log)
        self._tools_dock_widget.setWidget(self._control_panel_widget)

        self._connect_calibration_updates()

        log_view_dock_widget = QDockWidget("Log View", self)
        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, log_view_dock_widget)
        log_view_dock_widget.setWidget(self._log_view_widget)
        log_view_dock_widget.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable | QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )
        logger.debug("Finished initializing FreeMoCap MainWindow")

    def _create_tools_dock_widget(self):
        tools_dock_widget = QDockWidget("Control Panel", self)
        tools_dock_widget.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea | Qt.DockWidgetArea.RightDockWidgetArea)
        tools_dock_widget.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable | QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )
        return tools_dock_widget

    def _handle_videos_saved_to_this_folder_signal(self, folder_path: str):
        logger.debug(f"Videos saved to this folder signal received: {folder_path}")

        self._active_recording_info_widget.set_active_recording(recording_folder_path=folder_path)

        if (
                self._controller_group_box.auto_process_videos_checked
                and self._controller_group_box.mocap_videos_radio_button_checked
        ):
            logger.info("'Auto process videos' checkbox is checked - triggering 'Process Motion Capture Data' button")
            self._process_motion_capture_data_panel.process_motion_capture_data_button.click()
        elif self._controller_group_box.calibration_videos_radio_button_checked:
            logger.info("Processing calibration videos")
            self._process_motion_capture_data_panel.calibrate_from_active_recording(
                charuco_square_size_mm=float(self._controller_group_box.charuco_square_size),
                use_charuco_as_groundplane=self._controller_group_box.use_charuco_as_groundplane,
                charuco_board_name=self._controller_group_box.charuco_board_name,
            )

    def _handle_processing_finished_signal(self):
        self._update_skelly_viewer_widget()
        if self._controller_group_box.auto_open_in_blender_checked and not self._kill_thread_event.is_set():
            logger.info("'Auto Open in Blender' checkbox is checked - triggering 'Create Blender Scene'")
            self._export_active_recording_to_blender()
        if self._controller_group_box.generate_jupyter_notebook_checked and not self._kill_thread_event.is_set():
            self._generate_jupyter_notebook()
        logger.info("Processing finished")

    def handle_start_new_session_action(self):
        # self._central_tab_widget.set_welcome_tab_enabled(True)
        self._central_tab_widget.set_camera_view_tab_enabled(True)
        self._central_tab_widget.setCurrentIndex(1)
        self._controller_group_box.show()
        self._skellycam_widget.detect_available_cameras()

    def update(self):
        super().update()

        try:
            if not self._skellycam_widget.is_recording:
                self._controller_group_box.update_recording_name_string()
        except Exception as e:
            logger.exception(e)

    def _set_up_stylesheet(self):
        apply_css_style_sheet(self, get_css_stylesheet_path())
        SCSSFileWatcher(
            path_to_scss_file=get_scss_stylesheet_path(), path_to_css_file=get_css_stylesheet_path(), parent=self
        )
        css_file_watcher = CSSFileWatcher(path_to_css_file=get_css_stylesheet_path(), parent=self)
        return css_file_watcher

    def _create_new_synchronized_videos_folder(self) -> str:
        new_recording_folder_path = self._controller_group_box.get_new_recording_path()
        logger.info(f"Creating new recording folder at: {new_recording_folder_path}")
        self._active_recording_info_widget.set_active_recording(recording_folder_path=new_recording_folder_path)
        return self._active_recording_info_widget.active_recording_info.synchronized_videos_folder_path

    def _create_central_tab_widget(self):
        self._home_widget = HomeWidget(actions=self._actions, gui_state=self._gui_state, parent=self)

        self._skellycam_widget = SkellyCamWidget(
            self._create_new_synchronized_videos_folder,
            parent=self,
        )
        self._skellycam_widget.videos_saved_to_this_folder_signal.connect(
            self._handle_videos_saved_to_this_folder_signal
        )

        self._controller_group_box = CameraControllerGroupBox(
            skellycam_widget=self._skellycam_widget, gui_state=self._gui_state, parent=self
        )

        self._skelly_viewer_widget = SkellyViewer()

        center_tab_widget = CentralTabWidget(
            parent=self,
            skelly_cam_widget=self._skellycam_widget,
            camera_controller_widget=self._controller_group_box,
            welcome_to_freemocap_widget=self._home_widget,
            skelly_viewer_widget=self._skelly_viewer_widget,
            directory_view_widget=self._directory_view_widget,
            active_recording_info_widget=self._active_recording_info_widget,
        )

        center_tab_widget.set_welcome_tab_enabled(True)
        center_tab_widget.set_camera_view_tab_enabled(True)
        center_tab_widget.set_visualize_data_tab_enabled(True)

        return center_tab_widget

    def _create_directory_view_widget(self):
        return DirectoryViewWidget(
            gui_state=self._gui_state,
            get_active_recording_info_callable=self._active_recording_info_widget.get_active_recording_info,
        )

    def _create_control_panel_widget(self, log_update: Callable):
        self._camera_configuration_parameter_tree_widget = SkellyCamParameterTreeWidget(self._skellycam_widget)

        self._process_motion_capture_data_panel = ProcessMotionCaptureDataPanel(
            recording_processing_parameters=ProcessingParameterModel(),
            get_active_recording_info=self._active_recording_info_widget.get_active_recording_info,
            gui_state=self._gui_state,
            kill_thread_event=self._kill_thread_event,
            log_update=log_update,
        )
        self._process_motion_capture_data_panel.processing_finished_signal.connect(
            self._handle_processing_finished_signal
        )

        self._visualization_control_panel = VisualizationControlPanel(parent=self, gui_state=self._gui_state)
        self._visualization_control_panel.export_to_blender_button.clicked.connect(
            self._export_active_recording_to_blender
        )

        self._visualization_control_panel.generate_jupyter_notebook_button.clicked.connect(
            self._generate_jupyter_notebook
        )

        return ControlPanelWidget(
            camera_configuration_parameter_tree_widget=self._camera_configuration_parameter_tree_widget,
            process_motion_capture_data_panel=self._process_motion_capture_data_panel,
            visualize_data_widget=self._visualization_control_panel,
            parent=self,
        )

    def _export_active_recording_to_blender(self):
        logger.debug("Exporting active recording to Blender...")
        recording_path = self._active_recording_info_widget.get_active_recording_path()

        if self._visualization_control_panel.blender_executable_path is None:
            logger.error("Blender executable path is None!")
            return

        if not recording_path:
            logger.error("Recording path is None!")
            return

        self._export_to_blender_thread_worker = ExportToBlenderThreadWorker(
            recording_path=recording_path,
            blender_file_path=Path(get_blender_file_path(recording_path)),
            blender_executable_path=Path(self._visualization_control_panel.blender_executable_path),
            kill_thread_event=self._kill_thread_event,
        )
        self._export_to_blender_thread_worker.start()
        self._export_to_blender_thread_worker.success.connect(self._handle_export_to_blender_finished)

    @Slot()
    def _handle_export_to_blender_finished(self, success_value: bool) -> None:
        if success_value is False:
            logger.error("Blender export failed!")
        elif self._controller_group_box.auto_open_in_blender_checked:
            if Path(self._active_recording_info_widget.active_recording_info.blender_file_path).exists():
                open_file(self._active_recording_info_widget.active_recording_info.blender_file_path)
            else:
                logger.error(
                    "Blender file does not exist! Did something go wrong in the `export_to_blender` call above?"
                )

    def _generate_jupyter_notebook(self):
        logger.info("Exporting active recording to a Jupyter notebook...")
        recording_path = self._active_recording_info_widget.get_active_recording_path()
        # TODO: Need to include jupyter notebook in recording files that we keep track of (2023-05-15)
        if recording_path:
            generate_jupyter_notebook(path_to_recording=recording_path)

    def _handle_new_active_recording_selected(self, recording_info_model: RecordingInfoModel):
        logger.info(f"New active recording selected: {recording_info_model.path}")

        self._pipedream_pings.update_pings_dict(
            key=recording_info_model.name, value=self._active_recording_info_widget.active_recording_info.status_check
        )

        path = Path(recording_info_model.path)
        path.mkdir(parents=True, exist_ok=True)

        if str(path.parent) != get_recording_session_folder_path():
            self._directory_view_widget.set_folder_as_root(path.parent)
        else:
            self._directory_view_widget.set_folder_as_root(path)

        if Path(recording_info_model.synchronized_videos_folder_path).exists():
            self._directory_view_widget.expand_directory_to_path(recording_info_model.synchronized_videos_folder_path)
        else:
            self._directory_view_widget.expand_directory_to_path(recording_info_model.path)

        self._active_recording_info_widget.update_parameter_tree()
        # self._recording_name_label.setText(f"Recording Name: {recording_info_model.name}")
        self._update_skelly_viewer_widget()
        self._directory_view_widget.handle_new_active_recording_selected()

        try:
            self._process_motion_capture_data_panel.update_calibration_path()
        except (
                AttributeError
        ):  # Active Recording and Data Panel widgets rely on each other, so we're guaranteed to hit this every time the app opens
            logger.debug("Process motion capture data panel not created yet, skipping claibraiton setting")
        except Exception as e:
            logger.error(e)

        update_most_recent_recording_toml(recording_info_model=recording_info_model)

    def _update_skelly_viewer_widget(self):
        active_recording_info = self._active_recording_info_widget.active_recording_info

        if active_recording_info.data3d_status_check:
            self._skelly_viewer_widget.load_skeleton_data(
                mediapipe_skeleton_npy_path=active_recording_info.data_3d_npy_file_path
            )

        if active_recording_info.data2d_status_check:
            self._skelly_viewer_widget.generate_video_display(
                video_folder_path=active_recording_info.annotated_videos_folder_path
            )

        elif active_recording_info.synchronized_videos_status_check:
            self._skelly_viewer_widget.generate_video_display(
                video_folder_path=active_recording_info.synchronized_videos_folder_path
            )

    def kill_running_threads_and_processes(self):
        logger.info("Killing running threads and processes... ")
        try:
            self._skellycam_widget.close()
        except Exception as e:
            logger.error(f"Error killing running threads and processes: {e}")

        self._kill_thread_event.set()

    def handle_load_most_recent_recording(self):
        logger.info("`Load Most Recent Recording` QAction triggered")
        most_recent_recording_path = get_most_recent_recording_path()

        if most_recent_recording_path is None:
            logger.error("`get_most_recent_recording_path()` return `None`!")
            return

        self._active_recording_info_widget.set_active_recording(recording_folder_path=get_most_recent_recording_path())
        self._central_tab_widget.setCurrentIndex(2)
        self._control_panel_widget.tab_widget.setCurrentWidget(self._process_motion_capture_data_panel)

    def open_load_existing_recording_dialog(self):
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        logger.info("Opening `Load Existing Recording` dialog... ")
        user_selected_directory = QFileDialog.getExistingDirectory(
            self,
            "Select a recording folder",
            str(get_recording_session_folder_path()),
        )
        if len(user_selected_directory) == 0:
            logger.info("User cancelled `Load Existing Recording` dialog")
            return

        logger.info(f"User selected recording path:{user_selected_directory}")

        self._active_recording_info_widget.set_active_recording(recording_folder_path=user_selected_directory)
        self._central_tab_widget.setCurrentIndex(2)

    def reset_to_default_gui_settings(self):
        self._gui_state = GuiState()

        self._home_widget._send_pings_checkbox.setChecked(self._gui_state.send_user_pings)
        self._controller_group_box._auto_process_videos_checkbox.setChecked(self._gui_state.auto_process_videos_on_save)
        self._controller_group_box._generate_jupyter_notebook_checkbox.setChecked(
            self._gui_state.generate_jupyter_notebook
        )
        self._controller_group_box._auto_open_in_blender_checkbox.setChecked(self._gui_state.auto_open_in_blender)
        self._controller_group_box._charuco_square_size_line_edit.setText(str(self._gui_state.charuco_square_size))
        self._process_motion_capture_data_panel._calibration_control_panel._charuco_square_size_line_edit.setText(
            str(self._gui_state.charuco_square_size)
        )
        self._visualization_control_panel._blender_executable_label.setText(str(self._gui_state.blender_path))
        self._visualization_control_panel._blender_executable_path = str(self._gui_state.blender_path)

        save_gui_state(self._gui_state)

        self._active_recording_info_widget.set_active_recording(recording_folder_path=get_most_recent_recording_path())

    def open_import_videos_dialog(self):
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        logger.info("Opening `Import Videos` dialog... ")

        import_videos_path = QFileDialog.getExistingDirectory(
            self,
            "Select a folder containing synchronized videos (each video must have *exactly* the same number of frames)",
            str(Path.home()),
        )

        if len(import_videos_path) == 0:
            logger.info("User cancelled `Import Videos` dialog")
            return

        self._import_videos_window = ImportVideosWizard(
            parent=self,
            import_videos_path=import_videos_path,
            kill_thread_event=self._kill_thread_event,
        )
        self._import_videos_window.folder_to_save_videos_to_selected.connect(self._handle_import_videos)
        self._import_videos_window.exec()

    def open_welcome_screen_dialog(self):
        logger.info("Opening `Welcome to Freemocap` dialog... ")

        self._welcome_screen_dialog = WelcomeScreenDialog(
            gui_state=self._gui_state, kill_thread_event=self._kill_thread_event, parent=self
        )

        self._welcome_screen_dialog.exec()

    def open_release_notes_popup(self):
        logger.info("Opening `Release Notes` dialog... ")

        self._gui_state.shown_latest_release_notes = True

        save_gui_state(gui_state=self._gui_state, file_pathstring=get_gui_state_json_path())

        dialog = TabbedReleaseNotesDialog(
            kill_thread_event=threading.Event(),
            gui_state=self._gui_state,
            dark_mode=True,  # Set to True for dark mode, False for light mode
        )
        dialog.exec()

    def open_opencv_conflict_dialog(self):
        self._opencv_conflict_dialog = OpencvConflictDialog(
            gui_state=self._gui_state, kill_thread_event=self._kill_thread_event, parent=self
        )

        self._opencv_conflict_dialog.exec()

    def open_settings_dialog(self):
        self._settings_dialog = SetDataFolderDialog(
            gui_state=self._gui_state, kill_thread_event=self._kill_thread_event, parent=self
        )

        self._settings_dialog.exec()

        if self._settings_dialog.result():
            self.reboot_gui()

    def download_data(self, dataset_name: str):
        logger.info("Downloading sample data")
        self.download_data_thread_worker = DownloadDataThreadWorker(dataset_name=dataset_name)
        self.download_data_thread_worker.start()
        self.download_data_thread_worker.finished.connect(self._handle_download_data_finished)

    @Slot(str)
    def _handle_download_data_finished(self, downloaded_data_path: str):
        logger.info("Setting downloaded data as active recording... ")
        self._active_recording_info_widget.set_active_recording(recording_folder_path=downloaded_data_path)

    @Slot(list, str, bool)
    def _handle_import_videos(self, video_paths: List[str], folder_to_save_videos: str, synchronization_bool: bool):
        folder_to_save_videos = Path(folder_to_save_videos)
        folder_to_save_videos.mkdir(parents=True, exist_ok=True)

        if len(video_paths) == 0:
            logger.error("No videos to import!")
            return

        if not synchronization_bool:
            for video_path in tqdm(
                    video_paths,
                    desc="Importing videos...",
                    colour=[255, 128, 0],
                    unit="video",
                    unit_scale=True,
                    leave=False,
            ):
                if not Path(video_path).exists():
                    logger.error(f"{video_path} does not exist!")
                    return

                destination_path = folder_to_save_videos / Path(video_path).name
                logger.info(f"Copying video from {video_path} to {destination_path}")

                shutil.copy(video_path, destination_path)

        timestamps_copied = copy_directory_if_contains_timestamps(
            source_dir=Path(video_paths[0]).parent, destination_dir=folder_to_save_videos
        )

        if timestamps_copied:
            logger.info(f"Copied timestamps from {Path(video_paths[0]).parent} to {folder_to_save_videos}")
        else:
            logger.info(f"No timestamps found in {Path(video_paths[0]).parent}")

        self._active_recording_info_widget.set_active_recording(
            recording_folder_path=Path(folder_to_save_videos).parent
        )

    def _connect_calibration_updates(self):
        self._process_motion_capture_data_panel._calibration_control_panel.control_panel_calibration_updated.connect(
            self._controller_group_box.charuco_option_updated
        )
        self._controller_group_box.controller_group_box_calibration_updated.connect(
            self._process_motion_capture_data_panel._calibration_control_panel.charuco_option_updated
        )

    def reboot_gui(self):
        logger.info("Rebooting GUI... ")
        get_qt_app().exit(EXIT_CODE_REBOOT)

    def closeEvent(self, a0) -> None:
        logger.info("Main window `closeEvent` detected")

        if self._home_widget.consent_to_send_usage_information:
            self._pipedream_pings.update_pings_dict(key="gui_closed", value=True)
            if self._active_recording_info_widget.active_recording_info is not None:
                self._pipedream_pings.update_pings_dict(
                    key="active recording status on close",
                    value=self._active_recording_info_widget.active_recording_info.status_check,
                )
            self._pipedream_pings.send_pipedream_ping()

        try:
            remove_empty_directories(get_recording_session_folder_path())
        except Exception as e:
            logger.error(f"Error while removing empty directories: {e}")

        try:
            self._skellycam_widget.close()
        except Exception as e:
            logger.error(f"Error while closing the viewer widget: {e}")
        super().closeEvent(a0)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    main_window = MainWindow(pipedream_pings=PipedreamPings())
    main_window.show()
    app.exec()
    for process in multiprocessing.active_children():
        logger.info(f"Terminating process: {process}")
        process.terminate()
    sys.exit()



import logging
from typing import Dict, List, Optional

import numpy as np
from pydantic import BaseModel, ConfigDict

from freemocap.data_layer.skeleton_models.marker_info import MarkerInfo
from freemocap.data_layer.skeleton_models.segments import Segment, Segments, SegmentAnthropometry

logger = logging.getLogger(__name__)


class Skeleton(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    markers: MarkerInfo
    num_tracked_points: int
    segments: Optional[Dict[str, Segment]] = None
    marker_data: Dict[str, np.ndarray] = {}
    virtual_marker_data: Dict[str, np.ndarray] = {}
    joint_hierarchy: Optional[Dict[str, List[str]]] = None
    center_of_mass_definitions: Optional[Dict[str, SegmentAnthropometry]] = None
    num_frames: Optional[int] = None

    def add_segments(self, segment_connections: Dict[str, Segment]) -> None:
        """
        Adds segment connection data to the skeleton model.

        Parameters:
        - segment_connections: A dictionary where each key is a segment name and its value is a dictionary
          with information about that segment (e.g., 'proximal', 'distal' marker names).
        """
        segments_model = Segments(
            markers=self.markers,
            segment_connections={name: segment for name, segment in segment_connections.items()},
        )
        self.segments = segments_model.segment_connections

    def add_joint_hierarchy(self, joint_hierarchy: Dict[str, List[str]]) -> None:
        """
        Adds joint hierarchy data to the skeleton model.

        Parameters:
        - joint_hierarchy: A dictionary with joint names as keys and lists of connected marker names as values.
        """
        for joint_name, joint_connections in joint_hierarchy.items():
            if joint_name not in self.markers.all_markers:
                raise ValueError(f"The joint {joint_name} is not in the list of markers or virtual markers.")
            for connected_marker in joint_connections:
                if connected_marker not in self.markers.all_markers:
                    raise ValueError(
                        f"The connected marker {connected_marker} for {joint_name} is not in the list of markers or virtual markers."
                    )

        self.joint_hierarchy = joint_hierarchy

    def add_center_of_mass_definitions(self, center_of_mass_definitions: Dict[str, Dict[str, float]]) -> None:
        """
        Adds anthropometric center of mass definitions to the skeleton model.

        Parameters:
        - center_of_mass_definitions: A dictionary containing segment mass percentages.
        """
        if self.segments is None:
            raise ValueError("Segments must be defined before center of mass definitions can be added.")
        for com_segment_name in center_of_mass_definitions.keys():
            if com_segment_name not in self.segments.keys():
                raise ValueError(f"Segment {com_segment_name} is not in the list of segments.")

        self.center_of_mass_definitions = {
            name: SegmentAnthropometry(**values) for name, values in center_of_mass_definitions.items()
        }

    def integrate_freemocap_3d_data(self, freemocap_3d_data: np.ndarray) -> None:
        self.num_frames = freemocap_3d_data.shape[0]
        num_markers_in_data = freemocap_3d_data.shape[1]
        original_marker_names_list = self.markers.original_marker_names

        if num_markers_in_data != self.num_tracked_points:
            raise ValueError(
                f"The number of markers in the 3D data ({num_markers_in_data}) does not match "
                f"the expected number of tracked points ({self.num_tracked_points})."
            )

        self.marker_data = {
            marker_name: freemocap_3d_data[:, i, :] for i, marker_name in enumerate(original_marker_names_list)
        }

        try:
            self.calculate_virtual_markers()
        except ValueError:
            logger.info(
                "Freemocap data integrated without virtual markers, as no virtual marker definition was provided"
            )

    def calculate_virtual_markers(self) -> None:
        if not self.marker_data:
            raise ValueError(
                "3d marker data must be integrated before calculating virtual markers. Run `integrate_freemocap_3d_data()` first."
            )

        if not self.markers.virtual_marker_definition:
            raise ValueError(
                "Virtual marker info must be defined before calculating virtual markers. Run `add_virtual_markers()` first."
            )

        for vm_name, vm_info in self.markers.virtual_marker_definition.virtual_markers.items():
            vm_positions = np.zeros((self.marker_data[next(iter(self.marker_data))].shape[0], 3))
            for marker_name, weight in zip(vm_info["marker_names"], vm_info["marker_weights"]):
                vm_positions += self.marker_data[str(marker_name)] * weight
            self.virtual_marker_data[vm_name] = vm_positions

        self.marker_data.update(self.virtual_marker_data)

    def get_segment_markers(self, segment_name: str) -> Dict[str, np.ndarray]:
        """Returns a dictionary with the positions of the proximal and distal markers for a segment."""
        if not self.segments:
            raise ValueError("Segments must be defined before getting segment markers.")

        if not self.trajectories:
            raise ValueError("Trajectories must be defined before getting segment markers.")

        segment = self.segments.get(segment_name)
        if not segment:
            raise ValueError(f"Segment '{segment_name}' is not defined in the skeleton.")

        proximal_trajectories = self.trajectories.get(segment.proximal)
        distal_trajectories = self.trajectories.get(segment.distal)

        return {"proximal": proximal_trajectories, "distal": distal_trajectories}

    @property
    def trajectories(self) -> Dict[str, np.ndarray]:
        return self.marker_data



import inspect
import subprocess
from importlib.metadata import distribution
from pathlib import Path

from ajc27_freemocap_blender_addon.main import ajc27_run_as_main_function

from freemocap.core_processes.export_data.blender_stuff.export_to_blender.methods.ajc_addon.install.bpy_install_addon import (
    INSTALL_ADDON_SCRIPT_PATH,
)

FREEMOCAP_BLENDER_ADDON_PACKAGE_NAME = "ajc27_freemocap_blender_addon"


def get_package_path(package_name: str):
    try:
        package_dist = distribution(package_name)
        package_path = package_dist.locate_file("")
        return package_path
    except Exception:
        print(f"{package_name} not found.")
        return None


def install_ajc_addon(blender_exe_path: str, ajc_addon_main_file_path: str):
    addon_root_directory = Path(ajc_addon_main_file_path).parent
    addon_name = Path(addon_root_directory).name

    # Define your addon's name and root directory
    subprocess_command = [
        blender_exe_path,
        "--background",
        "--python",
        INSTALL_ADDON_SCRIPT_PATH,
        "--",
        addon_root_directory,
        addon_name,
    ]
    # use CLI to install the addon passing in the addon name and zip file path
    subprocess.run(subprocess_command)


if __name__ == "__main__":
    from freemocap.core_processes.export_data.blender_stuff.get_best_guess_of_blender_path import (
        get_best_guess_of_blender_path,
    )

    ajc_addon_main_file_path = inspect.getfile(ajc27_run_as_main_function)

    blender_path_in = get_best_guess_of_blender_path()

    install_ajc_addon(
        blender_exe_path=blender_path_in, ajc_addon_main_file_path=inspect.getfile(ajc27_run_as_main_function)
    )
    print("Done!")



import logging
import traceback
from typing import Optional, Union

from fastapi import APIRouter
from jon_scratch.pupil_calibration_pipeline.qt_gl_laser_skeleton_visualizer import (
    QtGlLaserSkeletonVisualizer,
)
from pydantic import BaseModel, ConfigDict
from src.cameras.launch_camera_frame_loop import launch_camera_frame_loop
from src.cameras.multicam_manager.cv_camera_manager import OpenCVCameraManager
from src.config.home_dir import (
    create_session_id,
    get_session_folder_path,
    get_most_recent_session_id,
    create_session_folder,
)
from src.core_processor.mediapipe_skeleton_detector.mediapipe_skeleton_detector import (
    MediaPipeSkeletonDetector,
)
from src.pipelines.calibration_pipeline.calibration_pipeline_orchestrator import (
    CalibrationPipelineOrchestrator,
)
from src.pipelines.session_pipeline.session_pipeline_orchestrator import (
    SessionPipelineOrchestrator,
    load_mediapipe3d_skeleton_data,
    load_mediapipe2d_data,
)

logger = logging.getLogger(__name__)

session_router = APIRouter()


class TweakedPydanticBaseModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed = True)


class SessionCreateModel(BaseModel):
    user_session_tag_str: Optional[str]


class SessionResponse(BaseModel):
    session_id: str
    session_path: str


class SessionCalibrateModel(TweakedPydanticBaseModel):
    """
    session_id: str, ID for the session we're calibrating
    charuco_square_size:Union[int, float], the size of one of the black squares on the charuco board (preferably) in milimeters
    """

    session_id: str = None
    webcam_configs_dict: dict = None
    opencv_camera_manager: OpenCVCameraManager = None
    charuco_square_size: Union[int, float] = 39


class SessionRecordModel(TweakedPydanticBaseModel):
    session_id: str = None
    webcam_configs_dict: dict = None


class SessionIdModel(BaseModel):
    """ """

    session_id: str = None


@session_router.post("/session/create")
async def create_session(
        session_create_model: SessionCreateModel = SessionCreateModel(),
) -> SessionResponse:
    session_id = create_session_id()
    if session_create_model.user_session_tag_str is not None:
        session_id = session_id + session_create_model.user_session_tag_str

    create_session_folder(session_id)

    return SessionResponse(
        session_id=session_id, session_path=get_session_folder_path(session_id)
    )


@session_router.post("/session/calibrate")
def calibrate_session(
        session_calibrate_model: SessionCalibrateModel = SessionCalibrateModel(),
):
    """calibate capture volume - record synchronized videos (from all available camras wtih default parameters for now) and process with Anipose to produce a camera calibration (saved as a `.toml` file in the session folder"""

    session_id = session_calibrate_model.session_id
    if session_id is None or session_id == "string":
        session_id = get_most_recent_session_id()

    calibration_orchestrator = CalibrationPipelineOrchestrator(session_id)

    # calibration_orchestrator.record_videos(show_visualizer_gui=False,
    #                                        save_video_in_frame_loop=False,
    #                                        show_camera_views_in_windows=True,
    #                                        )

    launch_camera_frame_loop(
        session_id=session_id,
        webcam_configs_dict=session_calibrate_model.webcam_configs_dict,
        show_camera_views_in_windows=True,
        calibration_videos_bool=True,
        detect_charuco_in_image=True,
    )

    try:
        calibration_orchestrator.run_anipose_camera_calibration(
            charuco_square_size=session_calibrate_model.charuco_square_size,
            pin_camera_0_to_origin=True,
        )
    except:
        logger.error("Printing Traceback")
        traceback.print_exc()

    return calibration_orchestrator


@session_router.post("/session/record")
def record_session(session_record_model: SessionRecordModel = SessionRecordModel()):
    launch_camera_frame_loop(
        session_id=session_record_model.session_id,
        webcam_configs_dict=session_record_model.webcam_configs_dict,
        show_camera_views_in_windows=True,
        calibration_videos_bool=False,
        detect_charuco_in_image=True,
    )


@session_router.post("/session/mediapipe_track_skeletons_offline")
def mediapipe_track_2D_skeletons_offline(
        session_id_model: SessionIdModel = SessionIdModel(),
):
    if session_id_model.session_id is None or session_id_model.session_id == "string":
        this_session_id = get_most_recent_session_id()
        logger.info(f"loading most recent session:{this_session_id}")
    else:
        this_session_id = session_id_model.session_id

    logger.info(
        f"tracking 2D mediapipe skeletons in videos from session: {this_session_id}"
    )
    mediapipe_skeleton_detector = MediaPipeSkeletonDetector(this_session_id)
    mediapipe_skeleton_detector.process_session_folder()


@session_router.post("/session/reconstruct_mediapipe3d_offline")
def mediapipe_reconstruct_3D_skeletons_offline(session_id_model: SessionIdModel = None):
    if session_id_model is None or session_id_model.session_id == "string":
        session_id = get_most_recent_session_id()
    else:
        session_id = session_id_model.session_id

    this_session_orchestrator = SessionPipelineOrchestrator(session_id=session_id)
    this_session_orchestrator.anipose_camera_calibration_object = (
        CalibrationPipelineOrchestrator().load_calibration_from_session_id(session_id)
    )
    this_session_orchestrator.mediapipe2d_numCams_numFrames_numTrackedPoints_XY = (
        load_mediapipe2d_data(session_id)
    )
    this_session_orchestrator.reconstruct3d_from_2d_data_offline()


@session_router.post("/session/visualize_offline")
def visualize_session_offline(session_id_model: SessionIdModel = None):
    if session_id_model is None or session_id_model.session_id == "string":
        session_id = get_most_recent_session_id()
    else:
        session_id = session_id_model.session_id

    mediapipe3d_data_payload = load_mediapipe3d_skeleton_data(session_id)

    mediapipe3d_skeleton_nFrames_nTrajectories_xyz = (
        mediapipe3d_data_payload.data3d_numFrames_numTrackedPoints_XYZ
    )
    mediapipe3d_skeleton_nFrames_nTrajectories_reprojectionError = (
        mediapipe3d_data_payload.data3d_numFrames_numTrackedPoint_reprojectionError
    )

    qt_gl_laser_skeleton = QtGlLaserSkeletonVisualizer(
        mediapipe_skel_fr_mar_xyz=mediapipe3d_skeleton_nFrames_nTrajectories_xyz
    )
    qt_gl_laser_skeleton.start_animation()


if __name__ == "__main__":
    # # create_session
    # session_id_in = create_session_id('session_router_as_main')
    # session_id_model = SessionIdModel(session_id=session_id_in)
    #
    # # # #calibrate_session
    # session_calibrate_model_in = SessionCalibrateModel(session_id=session_id_in,
    #                                                    charuco_square_size=39)
    # calibrate_session(session_calibrate_model_in)
    #
    # # # record new session
    # session_record_model_in = SessionRecordModel(session_id=session_id_in,)
    # record_session(session_record_model_in)
    #
    # # #process_
    # mediapipe_track_2D_skeletons_offline()
    #
    # mediapipe_reconstruct_3D_skeletons_offline()

    # #visualize with PyQt/OpenGL
    visualize_session_offline()



import logging
from copy import copy
from pathlib import Path
from typing import Union, Callable

from PySide6.QtCore import Signal, Qt
from PySide6.QtWidgets import QLabel, QMenu, QTreeView, QVBoxLayout, QWidget, QPushButton, QFileSystemModel
from qtpy import QtGui

from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState
from freemocap.system.open_file import open_file
from freemocap.system.paths_and_filenames.path_getters import get_recording_session_folder_path

logger = logging.getLogger(__name__)


class DirectoryViewWidget(QWidget):
    new_active_recording_selected_signal = Signal(str)

    def __init__(self, gui_state: GuiState, get_active_recording_info_callable: Callable):
        self._root_folder = None
        logger.debug("Creating QtDirectoryViewWidget")
        super().__init__()
        self._minimum_width = 300
        self.setMinimumWidth(self._minimum_width)

        self._gui_state = gui_state
        self._get_active_recording_info_callable = get_active_recording_info_callable

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._file_system_model = QFileSystemModel()

        self._tree_view_widget = QTreeView()
        self._tree_view_widget.setHeaderHidden(False)

        self._layout.addWidget(self._tree_view_widget)

        self._tree_view_widget.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._tree_view_widget.customContextMenuRequested.connect(self._context_menu)
        self._tree_view_widget.doubleClicked.connect(self._open_file)

        self._tree_view_widget.setModel(self._file_system_model)

        self._tree_view_widget.setAlternatingRowColors(True)
        self._tree_view_widget.setColumnWidth(0, 250)

        self._path_label = QLabel(str(self._gui_state.freemocap_data_folder_path))

        if self._gui_state.freemocap_data_folder_path is not None:
            self.set_folder_as_root(self._gui_state.freemocap_data_folder_path)

        self._show_freemocap_data_folder_button = QPushButton("Show FreeMoCap Data Folder")
        self._show_freemocap_data_folder_button.clicked.connect(
            lambda: self.set_folder_as_root(self._gui_state.freemocap_data_folder_path)
        )

        self._layout.addWidget(self._show_freemocap_data_folder_button)
        self._layout.addWidget(self._path_label)

    def expand_directory_to_path(self, path: Union[str, Path], collapse_other_directories: bool = True):
        if collapse_other_directories:
            logger.debug("Collapsing other directories")
            self._tree_view_widget.collapseAll()
        logger.debug(f"Expanding directory at  path: {str(path)}")
        og_index = self._file_system_model.index(str(path))
        self._tree_view_widget.expand(og_index)

        parent_path = copy(path)
        while Path(self._file_system_model.rootPath()) in Path(parent_path).parents:
            parent_path = Path(parent_path).parent
            index = self._file_system_model.index(str(parent_path))
            logger.debug(f"Expanding parent directory at  path: {str(parent_path)}")
            self._tree_view_widget.expand(index)

        self._tree_view_widget.scrollTo(og_index)

    def set_folder_as_root(self, folder_path: Union[str, Path]):
        logger.debug(f"Setting root folder to {str(folder_path)}")
        self._root_folder = folder_path
        self._tree_view_widget.setWindowTitle(str(folder_path))
        self._file_system_model.setRootPath(str(folder_path))
        self._tree_view_widget.setRootIndex(self._file_system_model.index(str(folder_path)))
        self._tree_view_widget.setColumnWidth(0, int(self._minimum_width * 0.9))

        self._path_label.setText(str(folder_path))

    def _context_menu(self):
        menu = QMenu()
        open = menu.addAction("Open file")
        open.triggered.connect(self._open_file)

        set_as_active_recording = menu.addAction("Set as Active Recording folder")
        set_as_active_recording.triggered.connect(self._set_recording_as_active)

        go_to_parent_directory = menu.addAction("Go to parent directory")
        go_to_parent_directory.triggered.connect(self._go_to_parent_directory)

        cursor = QtGui.QCursor()
        menu.exec(cursor.pos())

    def _open_file(self):
        index = self._tree_view_widget.currentIndex()
        file_path = self._file_system_model.filePath(index)
        logger.info(f"Opening file from file_system_view_widget: {file_path}")
        open_file(file_path)

    def _set_recording_as_active(self):
        index = self._tree_view_widget.currentIndex()
        file_path = Path(self._file_system_model.filePath(index))
        logger.info(f"Setting {file_path} as 'Active Recording' folder")
        self.new_active_recording_selected_signal.emit(str(file_path))

    def _go_to_parent_directory(self):
        logger.debug(f"Setting parent directory as root: {Path(self._root_folder).parent}")
        self.set_folder_as_root(Path(self._root_folder).parent)
        self.expand_directory_to_path(self._root_folder, collapse_other_directories=False)

    def set_path_as_index(self, path: Union[str, Path]):
        logger.info(f"Setting current index to : {str(path)}")
        self._tree_view_widget.setCurrentIndex(self._file_system_model.index(str(path)))

    def handle_new_active_recording_selected(self) -> None:
        current_recording_info = self._get_active_recording_info_callable()
        if current_recording_info is None:
            # self.set_folder_as_root(self._top_level_folder_path)
            self._show_freemocap_data_folder_button.hide()
            self.expand_directory_to_path(get_recording_session_folder_path())
            return

        # self.set_folder_as_root(current_recording_info.path)
        self._tree_view_widget.setCurrentIndex(self._file_system_model.index(str(current_recording_info.path)))

        if current_recording_info.annotated_videos_folder_path is not None:
            self.expand_directory_to_path(current_recording_info.annotated_videos_folder_path)
        elif current_recording_info.synchronized_videos_folder_path is not None:
            self.expand_directory_to_path(current_recording_info.synchronized_videos_folder_path)
        else:
            self.expand_directory_to_path(current_recording_info.path)


if __name__ == "__main__":
    import sys

    from PySide6.QtWidgets import QApplication

    app = QApplication(sys.argv)
    gui_state = GuiState()
    directory_view_widget = DirectoryViewWidget(gui_state=gui_state)

    directory_view_widget.show()

    # index = directory_view_widget.expand_directory_to_path(Path.home() / ".atom")
    directory_view_widget.expand_directory_to_path(Path.home() / "Downloads")

    sys.exit(app.exec())



import sys


def run_simple(recording_path_input: str, blender_file_save_path_input: str):
    from ajc27_freemocap_blender_addon.main import ajc27_run_as_main_function

    ajc27_run_as_main_function(recording_path=str(recording_path_input),
                               blend_file_path=str(blender_file_save_path_input))


if __name__ == "__main__":
    print(f"\nRunning {__file__} as a subprocess to install the addon...\n")
    argv = sys.argv
    print(f"Received command line arguments: {argv}")
    argv = argv[argv.index("--") + 1:]
    recording_path_input = str(argv[0])
    blender_file_save_path_input = str(argv[1])
    run_simple(recording_path_input=recording_path_input, blender_file_save_path_input=blender_file_save_path_input)

    print("\nDone!\n")



import logging
from collections import namedtuple
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Union

import cv2
import numpy as np

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration import (
    freemocap_anipose,
)
from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.charuco_groundplane_utils import (
    compute_basis_vectors_of_new_reference,
    find_good_frame,
    skellyforge_data,
    CharucoVisibilityError,
    CharucoVelocityError
)
from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)
from freemocap.system.paths_and_filenames.path_getters import (
    create_camera_calibration_file_name,
    get_calibrations_folder_path,
    get_last_successful_calibration_toml_path,
)
from freemocap.utilities.get_video_paths import get_video_paths

logger = logging.getLogger(__name__)

Extrinsics = namedtuple("Extrinsics", ["rvecs", "tvecs"])


@dataclass
class GroundPlaneSuccess:
    success: bool
    error: str = None


class AniposeCameraCalibrator:
    def __init__(
            self,
            charuco_board_object: CharucoBoardDefinition,
            charuco_square_size: Union[int, float],
            calibration_videos_folder_path: Union[str, Path],
            progress_callback: Callable[[str], None] = lambda _: None

    ):
        self._charuco_board_object = charuco_board_object
        self._progress_callback = progress_callback

        if charuco_square_size == 1:
            warning_string = "Charuco square size is not set, so units of 3d reconstructed data will be in units of `however_long_the_black_edge_of_the_charuco_square_was`. Please input `charuco_square_size` in millimeters (or your preferred unity of length)"
            logger.warning(warning_string)
            self._progress_callback(warning_string)
        self._charuco_square_size = charuco_square_size
        self._calibration_videos_folder_path = Path(calibration_videos_folder_path)
        self._recording_folder_path = Path(self._calibration_videos_folder_path).parent
        self._get_video_paths()
        self._initialize_anipose_objects()

    def _get_video_paths(
            self,
    ):
        self._list_of_video_paths = get_video_paths(path_to_video_folder=Path(self._calibration_videos_folder_path))

    def _initialize_anipose_objects(self):
        list_of_camera_names = [this_video_path.stem for this_video_path in self._list_of_video_paths]
        self._anipose_camera_group_object = freemocap_anipose.CameraGroup.from_names(list_of_camera_names)

        # add metadata
        self._anipose_camera_group_object.metadata["charuco_square_size"] = self._charuco_square_size
        self._anipose_camera_group_object.metadata["charuco_board_object"] = str(self._charuco_board_object)
        self._anipose_camera_group_object.metadata["path_to_recorded_videos"] = str(
            self._calibration_videos_folder_path
        )
        self._anipose_camera_group_object.metadata["date_time_calibrated"] = str(np.datetime64("now"))

        self._anipose_charuco_board = freemocap_anipose.AniposeCharucoBoard(
            self._charuco_board_object.number_of_squares_width,
            self._charuco_board_object.number_of_squares_height,
            square_length=self._charuco_square_size,  # mm
            marker_length=self._charuco_square_size * 0.8,
            marker_bits=4,
            dict_size=250,
        )

    def calibrate_camera_capture_volume(self, pin_camera_0_to_origin: bool = False,
                                        use_charuco_as_groundplane: bool = False) -> tuple[Path, GroundPlaneSuccess]:
        video_paths_list_of_list_of_strings = [[str(this_path)] for this_path in self._list_of_video_paths]

        (
            error,
            charuco_frame_data,
            charuco_frame_numbers,
        ) = self._anipose_camera_group_object.calibrate_videos(
            video_paths_list_of_list_of_strings, self._anipose_charuco_board
        )
        success_str = "Anipose Calibration Successful!"
        logger.info(success_str)
        self._progress_callback(success_str)
        self._anipose_camera_group_object.metadata["groundplane_calibration"] = False

        if pin_camera_0_to_origin:
            self._anipose_camera_group_object = self.pin_camera_zero_to_origin(self._anipose_camera_group_object)
            groundplane_success = None

        if use_charuco_as_groundplane:
            self._anipose_camera_group_object, groundplane_success = self.set_charuco_board_as_groundplane(
                self._anipose_camera_group_object)
            if groundplane_success.success:
                self._anipose_camera_group_object.metadata["groundplane_calibration"] = True

        self.get_real_world_matrices(self._anipose_camera_group_object)

        calibration_toml_filename = create_camera_calibration_file_name(
            recording_name=self._calibration_videos_folder_path.parent.stem
        )

        calibration_folder_toml_path = Path(get_calibrations_folder_path()) / calibration_toml_filename

        self._anipose_camera_group_object.dump(calibration_folder_toml_path)
        logger.debug(
            f"anipose camera calibration data saved to Calibrations folder - {str(calibration_folder_toml_path)}"
        )

        recording_folder_toml_path = Path(self._recording_folder_path) / calibration_toml_filename

        self._anipose_camera_group_object.dump(recording_folder_toml_path)
        logger.debug(f"anipose camera calibration data saved to Recording folder - {str(recording_folder_toml_path)}")

        last_successful_calibration_toml_path = get_last_successful_calibration_toml_path()
        self._anipose_camera_group_object.dump(last_successful_calibration_toml_path)

        logger.debug(
            f"anipose camera calibration data also saved to 'Last Successful Calibration' - {last_successful_calibration_toml_path}"
        )
        self._progress_callback(
            "Anipose camera calibration data saved to calibrations folder, recording folder, and `Last Successful Calibration` file"
        )

        return calibration_folder_toml_path, groundplane_success

    def pin_camera_zero_to_origin(self, cam_group: freemocap_anipose.CameraGroup):
        """
        Re-express all camera extrinsics relative to camera 0.

        This function performs two operations:
            1. Rotates all camera coordinate systems so that camera 0's orientation becomes the new world frame.
            2. Shifts the world origin to camera 0's position, making it the new coordinate origin (0, 0, 0).

        The result is that:
            - Camera 0 ends up with an identity rotation and zero translation.
            - All other cameras are expressed relative to camera 0's original position and orientation.

        Args:
            cam_group (freemocap_anipose.CameraGroup): A group of calibrated cameras whose extrinsics will be adjusted in place.
        """

        rvecs_new = self.align_rotations_to_cam0(cam_group=cam_group)
        cam_group.set_rotations(rvecs_new)
        tvecs_new = self.shift_origin_to_cam0(cam_group=cam_group)
        cam_group.set_translations(tvecs_new)
        return cam_group

    def align_rotations_to_cam0(self, cam_group: freemocap_anipose.CameraGroup):
        rvecs = cam_group.get_rotations()
        # get rotation of cam 0 to world
        R0, _ = cv2.Rodrigues(rvecs[0])

        rvecs_new = np.empty_like(rvecs)
        # rotates each cameras coordinate system into the new world space
        # R0 maps world -> cam0, so R0.T (which is also the inverse) maps cam0 -> world
        # R0 * R0^-1 = identity matrix, so this makes R0 the origin
        # then we map every other camera into the new world space
        for i in range(rvecs.shape[0]):
            Ri, _ = cv2.Rodrigues(rvecs[i])
            Ri_new, _ = cv2.Rodrigues(Ri @ R0.T)
            rvecs_new[i] = Ri_new.flatten()
        return rvecs_new

    def shift_origin_to_cam0(self, cam_group: freemocap_anipose.CameraGroup):
        # Get original translation and rotation vectors
        tvecs = cam_group.get_translations()
        rvecs = cam_group.get_rotations()

        # Extract camera 0's rotation and translation
        camera_0_translation = tvecs[0, :]
        camera_0_rotation = rvecs[0, :]

        # Get 3x3 rotation matrix for world -> cam 0
        R0, _ = cv2.Rodrigues(camera_0_rotation)

        # Get the vector that moves the world origin to cam 0 in the world frame
        delta_to_origin_world = - R0.T @ camera_0_translation

        # Create new translation vectors array
        new_tvecs = np.zeros_like(tvecs)

        # Apply offset to each camera's translation vector
        for cam_i in range(tvecs.shape[0]):
            Ri, _ = cv2.Rodrigues(rvecs[cam_i, :])
            # Transform the world offset into this camera's coordinate frame
            delta_to_origin_camera_i = Ri @ delta_to_origin_world
            # Update the translation vector
            new_tvecs[cam_i, :] = tvecs[cam_i, :] + delta_to_origin_camera_i
        return new_tvecs

    def set_charuco_board_as_groundplane(self, cam_group: freemocap_anipose.CameraGroup):
        # Get the camera group object
        logger.info("Getting 2d Charuco data")

        charuco_2d_xy: np.ndarray = cam_group.charuco_2d_data

        if charuco_2d_xy is None:
            error_message = "Charuco 2d data was not retrieved successfully. Check for an error during calibration."
            logger.error(error_message)
            raise ValueError(error_message)
        logger.info("Charuco 2d data retrieved successfully with shape: "
                    f"{charuco_2d_xy.shape}")
        charuco_2d_xy = charuco_2d_xy.astype(np.float64)
        num_cameras, num_frames, num_tracked_points, _ = charuco_2d_xy.shape
        charuco_2d_flat = charuco_2d_xy.reshape(num_cameras, -1, 2)

        logger.info("Getting 3d Charuco data")
        charuco_3d_flat = self._anipose_camera_group_object.triangulate(
            charuco_2d_flat)
        charuco_3d_xyz = charuco_3d_flat.reshape(num_frames, num_tracked_points, 3)
        logger.info(f"Charuco 3d data reconstructed with shape: {charuco_3d_xyz.shape}")

        charuco_3d_xyz_interpolated = skellyforge_data(raw_charuco_data=charuco_3d_xyz)

        try:
            charuco_still_frame_idx = find_good_frame(charuco_data=charuco_3d_xyz_interpolated,
                                                      number_of_squares_width=self._charuco_board_object.number_of_squares_width,
                                                      number_of_squares_height=self._charuco_board_object.number_of_squares_height, )
        except CharucoVisibilityError as e:
            logger.warning(
                "Ground-plane alignment skipped – reverting to original calibration: %s",
                e,
                exc_info=True, )
            success = GroundPlaneSuccess(success=False, error=str(e))
            return cam_group, success
        except CharucoVelocityError as e:
            logger.warning(
                "Ground-plane alignment skipped – reverting to original calibration: %s",
                e,
                exc_info=True, )
            success = GroundPlaneSuccess(success=False, error=str(e))
            return cam_group, success

        charuco_frame = charuco_3d_xyz_interpolated[charuco_still_frame_idx]

        x_hat, y_hat, z_hat = compute_basis_vectors_of_new_reference(charuco_frame,
                                                                     number_of_squares_width=self._charuco_board_object.number_of_squares_width,
                                                                     number_of_squares_height=self._charuco_board_object.number_of_squares_height)
        charuco_origin_in_world = charuco_frame[0]
        rmat_charuco_to_world = np.column_stack([x_hat, y_hat,
                                                 z_hat])  # rotation matrix that transforms a point in the new world reference frame(defined by the charuco) into legacy world frame

        extrinsics = self.adjust_world_reference_frame_to_charuco(cam_group=cam_group,
                                                                  charuco_origin_in_world=charuco_origin_in_world,
                                                                  rmat_charuco_to_world=rmat_charuco_to_world)
        cam_group.set_rotations(extrinsics.rvecs)
        cam_group.set_translations(extrinsics.tvecs)
        success = GroundPlaneSuccess(success=True)
        logger.info("Anipose camera calibration data adjusted to set charuco board as ground plane")

        charuco_3d_flat = self._anipose_camera_group_object.triangulate(
            charuco_2d_flat)
        charuco_3d_xyz = charuco_3d_flat.reshape(num_frames, num_tracked_points, 3)
        logger.info(f"Charuco 3d data reconstructed with shape: {charuco_3d_xyz.shape}")

        charuco_3d_xyz_interpolated = skellyforge_data(raw_charuco_data=charuco_3d_xyz)

        charuco_save_path = self._recording_folder_path / "output_data" / "charuco_3d_xyz.npy"
        np.save(charuco_save_path,
                charuco_3d_xyz_interpolated)  # NOTE - there may be a more robust way to get this path but this is the simplest for now
        logger.info(f"Charuco 3d data saved to {charuco_save_path}")

        return cam_group, success

    def adjust_world_reference_frame_to_charuco(self,
                                                cam_group: freemocap_anipose.CameraGroup,
                                                charuco_origin_in_world: np.ndarray,
                                                rmat_charuco_to_world: np.ndarray):

        tvecs = cam_group.get_translations()
        rvecs = cam_group.get_rotations()

        tvecs_new = np.zeros_like(tvecs)
        rvecs_new = np.zeros_like(rvecs)

        for i in range(tvecs.shape[0]):
            rmat_world_to_cam_i, _ = cv2.Rodrigues(rvecs[i])
            # Xcam = Rc->w*Xworld + tvec (base formula for this transformation)
            t_delta = rmat_world_to_cam_i @ charuco_origin_in_world  # Rc->w*Xworld (rmat is Rc->w)
            tvecs_new[i] = t_delta + tvecs[i]  # Rc->w*Xworld + tvec

            new_rmat = rmat_world_to_cam_i @ rmat_charuco_to_world  # composes w' -> w (rotation matrix) with w -> c (rmat) to get w' -> c
            new_rvec, _ = cv2.Rodrigues(new_rmat)
            rvecs_new[i] = new_rvec.flatten()
        return Extrinsics(rvecs_new, tvecs_new)

    def get_real_world_matrices(self,
                                cam_group: freemocap_anipose.CameraGroup):
        rvecs = cam_group.get_rotations()
        tvecs = cam_group.get_translations()

        positions = []
        orientations = []
        for i in range(tvecs.shape[0]):
            rmat_world_to_cam_i, _ = cv2.Rodrigues(rvecs[i])

            rmat_cam_to_world = rmat_world_to_cam_i.T
            t_world = -rmat_cam_to_world @ tvecs[i]
            positions.append(t_world.astype(float).tolist())
            orientations.append(rmat_cam_to_world.astype(float).tolist())

        cam_group.set_world_positions(positions)
        cam_group.set_world_orientations(orientations)

        return positions, orientations



from pathlib import Path
from typing import Union

import sass


def compile_scss_to_css(scss_path: Union[str, Path], css_path: Union[str, Path]):
    with open(scss_path) as scss_file:
        scss_contents = scss_file.read()

    compiled_css = sass.compile(string=scss_contents)

    with open(css_path, "w") as css_file:
        css_file.write(compiled_css)



from pathlib import Path
import numpy as np
from time import perf_counter_ns

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.freemocap_anipose import CameraGroup


def direct_triangulate(anipose_calibration_object: CameraGroup, data_2d: np.ndarray):
    return anipose_calibration_object.triangulate(data_2d, progress=False)


if __name__ == "__main__":
    # TODO: do all this setup with recording_info object
    recording_folder = Path("/Users/philipqueen/freemocap_data/recording_sessions/freemocap_test_data/")

    calibration_toml_path = list(recording_folder.glob("*calibration.toml"))[0]
    calibration = CameraGroup.load(calibration_toml_path)

    data_2d_path = (
        recording_folder
        / "output_data"
        / "raw_data"
        / "mediapipe_2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy"
    )
    data_2d = np.load(data_2d_path)
    data_2d_flat = data_2d.reshape(data_2d.shape[0], -1, 2)

    num_trials = 5

    times = []
    for i in range(num_trials):
        start = perf_counter_ns()
        direct_triangulate(calibration, data_2d_flat)
        end = perf_counter_ns()
        times.append((end - start) / 1e9)
        print(f"This round triangulation time is: {(end - start) / 1e9} seconds for {data_2d.shape[1]} frames")
        print(f"This round average time per frame is: {(end - start) / 1e9 / data_2d.shape[1]} seconds\n")
    print(
        f"\tTotal triangulation time is: {np.mean(times)} seconds for {data_2d.shape[1]} frames over {num_trials} runs"
    )
    print(f"\tAverage time per frame is: {np.mean(times) / (data_2d.shape[1] * num_trials)} seconds")



from pathlib import Path
from typing import Union, List


def get_video_paths(path_to_video_folder: Union[str, Path]) -> List[Path]:
    """Search the folder for 'mp4' files (case insensitive) and return them as a list"""

    list_of_video_paths = list(Path(path_to_video_folder).glob("*.mp4")) + list(
        Path(path_to_video_folder).glob("*.MP4")
    )
    unique_list_of_video_paths = get_unique_list(list_of_video_paths)
    return sorted(unique_list_of_video_paths, key=lambda p: str(p).lower())


def get_unique_list(list: list) -> list:
    """Return a list of the unique elements from input list"""
    unique_list = []
    [unique_list.append(element) for element in list if element not in unique_list]

    return unique_list



import logging
from pathlib import Path
from typing import Optional, Union
from multiprocessing import cpu_count

from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import MediapipeTrackingParams

from freemocap.core_processes.export_data.blender_stuff.export_to_blender.export_to_blender import export_to_blender
from freemocap.core_processes.process_motion_capture_videos.process_recording_folder import process_recording_folder
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel
from freemocap.data_layer.recording_models.recording_info_model import RecordingInfoModel
from freemocap.system.paths_and_filenames.file_and_folder_names import RECORDING_PARAMETERS_JSON_FILE_NAME
from freemocap.system.paths_and_filenames.path_getters import get_blender_file_path
from freemocap.utilities.save_dictionary_to_json import save_dictionary_to_json


def process_folder_of_session_folders(
    path_to_folder_of_session_folders: Union[str, Path],
    path_to_camera_calibration_toml: Optional[Union[str, Path]] = None,
    path_to_blender_executable: Optional[Union[str, Path]] = None,
    max_num_processes: int = cpu_count() - 1,
):
    # list_of_session_folders = list(path_to_folder_of_session_folders.glob("ses*"))
    list_of_session_folders = [f for f in path_to_folder_of_session_folders.iterdir() if f.is_dir()]
    print(list_of_session_folders)

    for session_folder_path in list_of_session_folders:
        logging.info(f"Looking in {session_folder_path} for recording folders")
        list_of_recording_folders = [
            f
            for f in session_folder_path.iterdir()
            if f.is_dir() and (f / "synchronized_videos").exists() and any((f / "synchronized_videos").iterdir())
        ]

        calibration_files = [
            toml_file for sesh in list_of_recording_folders for toml_file in list(sesh.glob("*calibration.toml"))
        ]

        for recording_folder_path in list_of_recording_folders:
            logging.info("Looking for calibration toml")
            if calibration_files:
                this_recording_calibration_toml_path = calibration_files[0]
                logging.info(f"Using {this_recording_calibration_toml_path} for this recording")
            else:
                this_recording_calibration_toml_path = path_to_camera_calibration_toml
                logging.info("Using manually entered calibration for this recording")

            logging.info(f"Processing {recording_folder_path}")

            process_recording_without_gui(
                recording_path=recording_folder_path,
                path_to_blender_executable=path_to_blender_executable,
                path_to_camera_calibration_toml=this_recording_calibration_toml_path,
                max_num_processes=max_num_processes,
            )

    logging.info("Done!")


def process_recording_without_gui(
    recording_path: Union[str, Path],
    path_to_camera_calibration_toml: Optional[Union[str, Path]] = None,
    path_to_blender_executable: Optional[Union[str, Path]] = None,
    max_num_processes: int = cpu_count() - 1,
):
    rec = ProcessingParameterModel(
        tracking_parameters_model=MediapipeTrackingParams(num_processes=max_num_processes),
    )

    rec.recording_info_model = RecordingInfoModel(recording_folder_path=Path(recording_path))

    if path_to_camera_calibration_toml:
        rec.recording_info_model.calibration_toml_path = Path(path_to_camera_calibration_toml)
    else:
        logging.warning("No camera calibration toml file provided. May cause an error with multicamera recordings.")

    recording_info_dict = rec.model_dump(exclude={"recording_info_model"})

    Path(rec.recording_info_model.output_data_folder_path).mkdir(parents=True, exist_ok=True)

    save_dictionary_to_json(
        save_path=rec.recording_info_model.output_data_folder_path,
        file_name=RECORDING_PARAMETERS_JSON_FILE_NAME,
        dictionary=recording_info_dict,
    )

    process_recording_folder(recording_processing_parameter_model=rec)

    if path_to_blender_executable:
        blender_file_path = get_blender_file_path(recording_folder_path=recording_path)
        logging.info(f"Exporting to {blender_file_path}")
        export_to_blender(
            recording_folder_path=recording_path,
            blender_file_path=blender_file_path,
            blender_exe_path=Path(path_to_blender_executable),
        )
    else:
        logging.warning("No blender executable provided. Blender file will not be exported.")


if __name__ == "__main__":
    path_to_folder_of_session_folders = Path(r"D:\footropter_pilot_04_19_23\1.0_recordings")

    # Add path to camera calibration toml file you would like to default to
    path_to_camera_calibration_toml = None

    path_to_blender_executable = Path(r"C:\Program Files\Blender Foundation\Blender 3.1\blender.exe")

    process_folder_of_session_folders(
        path_to_folder_of_session_folders=path_to_folder_of_session_folders,
        path_to_camera_calibration_toml=path_to_camera_calibration_toml,
        path_to_blender_executable=path_to_blender_executable,
    )



import uvicorn
from fastapi import FastAPI

from src.api.middleware.cors import cors
from src.api.routes import enabled_routers


def create_app(*args, **kwargs):
    _app = FastAPI()

    cors(_app)

    for router in enabled_routers:
        _app.include_router(router)

    return _app


if __name__ == "__main__":
    script_app = create_app()
    uvicorn.run(script_app, host="127.0.0.1", port=8080, log_level="info")



from pathlib import Path
from typing import Union

import numpy as np
import pytest

from freemocap.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)
from freemocap.utilities.get_video_paths import get_video_paths


@pytest.mark.usefixtures("synchronized_video_folder_path", "image_tracking_data")
def test_image_tracking_data_exists(
        synchronized_video_folder_path: Union[str, Path],
        image_tracking_data: Union[str, Path],
):
    """
    test that the `2d detection` process worked correctly by checking that there is an `.npy` file containing the 2d data in the `output_data_folder`
    """

    assert Path(image_tracking_data).is_file(), f"{image_tracking_data} is not a file"


@pytest.mark.usefixtures("synchronized_video_folder_path", "image_tracking_data")
def test_image_tracking_data_shape(
        synchronized_video_folder_path: Union[str, Path],
        image_tracking_data: Union[str, Path, np.ndarray],
):
    """
    test that the `2d detection` process worked correctly by checking:

    The dimensions of that `npy` (number of cameras, number of frames, [ need to do - number of tracked points], [pixelX, pixelY] matches the videos in the `synchronized videos` folder

    TODO - check number of tracked points vs 'expected' number of tracked points
    """
    if isinstance(image_tracking_data, (str, Path)):
        image2d_camera_frame_marker_xyz = np.load(image_tracking_data)
    else:
        image2d_camera_frame_marker_xyz = image_tracking_data

    list_of_video_paths = get_video_paths(Path(synchronized_video_folder_path))
    number_of_videos = len(list_of_video_paths)
    assert (
            image2d_camera_frame_marker_xyz.shape[0] == number_of_videos
    ), f"Number of videos in image_tracking_data ({image2d_camera_frame_marker_xyz.shape[0]}) does not match number of videos in synchronized videos folder ({number_of_videos})"

    frame_counts = list(get_number_of_frames_of_videos_in_a_folder(synchronized_video_folder_path).values())

    assert (
            len(set(frame_counts)) == 1
    ), f"Videos in {synchronized_video_folder_path} have different frame counts: {frame_counts}"
    assert (
            image2d_camera_frame_marker_xyz.shape[1] == frame_counts[0]
    ), f"Number of frames in image_tracking_data does not match number of frames of videos in {synchronized_video_folder_path}"

    # TODO - check number of tracked points vs 'expected' number of tracked points

    assert (
            image2d_camera_frame_marker_xyz.shape[3] == 3
    ), f"Data has {image2d_camera_frame_marker_xyz.shape[3]} dimensions, expected 3 dimensions"



import os
import shutil
import tempfile
from pathlib import Path
from typing import Union

INSTALL_ADDON_SCRIPT_PATH = __file__


def create_zip_from_folder(addon_folder_path: Union[Path, str]) -> str:
    try:
        addon_folder_path = Path(addon_folder_path)
        if not addon_folder_path.is_dir():
            raise FileNotFoundError(f"Addon folder path `{addon_folder_path}` does not exist!")
        # if not addon_folder_path.joinpath("__init__.py").is_file():
        #     raise FileNotFoundError(f"Addon folder path `{addon_folder_path}` does not contain an `__init__.py` file!")

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir) / addon_folder_path.name
            shutil.copytree(addon_folder_path, tmp_path)
            zip_path = addon_folder_path.parent / f"{addon_folder_path.name}.zip"
            shutil.make_archive(base_name=str(zip_path.with_suffix("")), format="zip", root_dir=tmp_path.parent)

        if not Path(zip_path).is_file():
            raise FileNotFoundError(f"Failed to create zip file at : {zip_path}")
    except Exception as e:
        print("Failed to create zip file at")
        print(e)
        raise e
    return str(zip_path)


def bpy_install_addon(addon_root_directory: str):
    """
    This gets called as a subprocess to install the addon to blender
    """
    import bpy

    addon_root_directory = Path(addon_root_directory)

    addon_name = Path(addon_root_directory).name

    print(f"\n\nInstalling blender addon named `{addon_name}` from root directory: `{addon_root_directory}`\n\n")
    zip_file_path = create_zip_from_folder(addon_folder_path=addon_root_directory)
    print(f"Created zip file at `{zip_file_path}`\n\n")

    print("\n\nInstalling addon from zip file...\n\n")
    bpy.ops.preferences.addon_install(overwrite=True, target="DEFAULT", filepath=zip_file_path)

    print("\n\nSaving user preferences...\n\n")
    bpy.ops.wm.save_userpref()
    # print(f"\n\nEnabling addon `{addon_name}`...\n\n")
    # bpy.ops.preferences.addon_enable(module=f'{addon_name}')
    #
    # installed_addons = bpy.context.preferences.addons.keys()
    # # Check if the addon name is in the list of installed addons
    #
    # if addon_name not in (bpy.context.preferences.addons.keys()):
    #     raise Exception(f"Failed to install `{addon_name}`!!")
    # else:
    #     print("SUCCESSFULLY INSTALLED `{addon_name}`!!")

    # Remove the temporary zipfile
    os.remove(zip_file_path)


if __name__ == "__main__":
    print(f"Running {__file__} as a subprocess to install the addon...")
    import sys

    # take in cli args split by --, so this is the first arg
    argument_variables = sys.argv
    argv = sys.argv[sys.argv.index("--") + 1:]
    addon_root_directory = argv[0]
    addon_name = argv[1]
    if not Path(addon_root_directory).exists():
        raise FileNotFoundError(f"Addon root directory `{addon_root_directory}` does not exist!")
    bpy_install_addon(addon_root_directory=addon_root_directory)



import numpy as np


def project_3d_data_to_z_plane(skel3d_frame_marker_xyz: np.ndarray) -> np.ndarray:
    projected_skel3d_frame_marker_xyz = np.zeros(skel3d_frame_marker_xyz.shape)

    projected_skel3d_frame_marker_xyz[:, :, 0] = skel3d_frame_marker_xyz[:, :, 0]
    projected_skel3d_frame_marker_xyz[:, :, 1] = skel3d_frame_marker_xyz[:, :, 1]
    projected_skel3d_frame_marker_xyz[:, :, 2] = 0

    return projected_skel3d_frame_marker_xyz



import logging
from typing import Union

import requests
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
)
from packaging import version

import freemocap
from freemocap.gui.qt.actions_and_menus.actions import (
    CREATE_NEW_RECORDING_ACTION_NAME,
    LOAD_RECORDING_ACTION_NAME,
    IMPORT_VIDEOS_ACTION_NAME,
    Actions,
)
from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState, save_gui_state
from freemocap.gui.qt.widgets.logo_svg_widget import LogoSvgWidget
from freemocap.system.paths_and_filenames.file_and_folder_names import DOCUMENTATION_HOME, \
    DOCUMENTATION_PRIVACY_POLICY_URL, PATH_TO_FREEMOCAP_LOGO_SVG
from freemocap.system.paths_and_filenames.path_getters import get_gui_state_json_path

logger = logging.getLogger(__name__)


class WelcomeScreenButton(QPushButton):
    def __init__(self, text: str, parent: QWidget = None):
        super().__init__(text, parent=parent)
        self.setFixedHeight(50)
        self.setFixedWidth(400)


class HomeWidget(QWidget):
    def __init__(self, actions: Actions, gui_state: GuiState, parent: QWidget = None):
        super().__init__(parent=parent)

        self.gui_state = gui_state

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._layout.setContentsMargins(0, 0, 0, 0)

        self.sizePolicy().setHorizontalStretch(1)
        self.sizePolicy().setVerticalStretch(1)
        self._layout.addStretch(1)
        self._add_freemocap_logo()

        self._welcome_to_freemocap_title_widget = self._welcome_to_freemocap_title()
        self._layout.addWidget(self._welcome_to_freemocap_title_widget)

        self._create_new_session_button = WelcomeScreenButton(f"{CREATE_NEW_RECORDING_ACTION_NAME}")
        self._create_new_session_button.clicked.connect(actions.create_new_recording_action.trigger)
        self._layout.addWidget(self._create_new_session_button, alignment=Qt.AlignmentFlag.AlignCenter)
        self._create_new_session_button.setProperty("recommended_next", True)

        self._load_existing_session_button = WelcomeScreenButton(f"{LOAD_RECORDING_ACTION_NAME}")
        self._load_existing_session_button.clicked.connect(actions.load_existing_recording_action.trigger)
        self._layout.addWidget(self._load_existing_session_button, alignment=Qt.AlignmentFlag.AlignCenter)

        self._import_videos_button = WelcomeScreenButton(f"{IMPORT_VIDEOS_ACTION_NAME}")
        self._import_videos_button.clicked.connect(actions.import_videos_action.trigger)
        self._layout.addWidget(self._import_videos_button, alignment=Qt.AlignmentFlag.AlignCenter)

        self._create_user_info_consent_checkbox()

        self._add_text_links()
        self._make_version_label()

        self.style().polish(self)

    def _make_version_label(self):
        hbox = QHBoxLayout()
        self._layout.addLayout(hbox)
        version_label_string = f'source:<a href="https://github.com/freemocap/freemocap" style="color: #777777;"> {freemocap.__version__}</a>'
        latest_version = self.check_for_latest_version()
        if latest_version is not None:
            latest_version = latest_version.split("rc")[0]
        current_version = freemocap.__version__.strip("v").split("-")[0]

        if latest_version is None:
            version_label_string += " (X)"
            tooltip_string = " (`new version check` failed!)"
        elif latest_version != current_version:
            version_label_string += " (update available!)"
            tooltip_string = (
                f"New version {latest_version} available!\n"
                f" upgrade to latest version by entering the command \n\n"
                f" ```\npip install freemocap --upgrade\n``` "
                f"\n\ninto the terminal you used to launch this program (with your `(freemocap-env)` environment activated)"
            )
        else:
            tooltip_string = f"Your version of freemocap ({freemocap.__version__}) is up to date! (latest: {self.check_for_latest_version()})"
        version_label = QLabel(version_label_string)
        version_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        version_label.setStyleSheet("font-size: 12px;color: #777777")
        version_label.setOpenExternalLinks(True)
        version_label.setToolTip(tooltip_string)

        hbox.addWidget(version_label)

    def check_for_latest_version(self) -> Union[str, None]:
        try:
            response = requests.get(f"https://pypi.org/pypi/{freemocap.__package_name__}/json", timeout=(5, 60))
        except requests.RequestException:
            logger.error(f"Failed to check for latest version of {freemocap.__package_name__}")
            return None

        versions = response.json()["releases"].keys()

        pre_releases = [v for v in versions if "a" in v or "b" in v or "rc" in v]
        if pre_releases:
            pre_releases.sort(key=version.parse)

        versions = list(versions)
        versions.sort(key=version.parse)
        all_versions = pre_releases + versions
        all_versions.sort(key=version.parse)

        return all_versions[-1] if all_versions else None

    def _add_text_links(self):
        hbox = QHBoxLayout()
        self._layout.addLayout(hbox)
        hbox.addStretch(1)
        privacy_policy_link_string = f'<a href="{DOCUMENTATION_PRIVACY_POLICY_URL}" style="color: #333333;">privacy policy</a>'
        privacy_policy_link_label = QLabel(privacy_policy_link_string)
        privacy_policy_link_label.setOpenExternalLinks(True)
        hbox.addWidget(privacy_policy_link_label)
        docs_string = f'<a href="{DOCUMENTATION_HOME}" style="color: #333333;">docs</a>'
        docs_string = QLabel(docs_string)
        docs_string.setOpenExternalLinks(True)
        hbox.addWidget(docs_string, alignment=Qt.AlignmentFlag.AlignCenter)
        feedback_string = '<a href="https://forms.gle/AguGxQAJGXxaJbXdA" style="color: #333333;">feedback</a>'
        feedback_string = QLabel(feedback_string)
        feedback_string.setOpenExternalLinks(True)
        hbox.addWidget(feedback_string, alignment=Qt.AlignmentFlag.AlignRight)
        donate_string = '<a href="https://freemocap.org/about-us.html#donate" style="color: #333333;">donate &lt;3</a>'
        donate_string = QLabel(donate_string)
        donate_string.setOpenExternalLinks(True)
        hbox.addWidget(donate_string, alignment=Qt.AlignmentFlag.AlignRight)

        hbox.addStretch(1)

    def _create_user_info_consent_checkbox(self):
        hbox = QHBoxLayout()
        self._layout.addLayout(hbox)
        hbox.addStretch(1)
        self._send_pings_checkbox = QCheckBox("Send anonymous usage information")
        self._send_pings_checkbox.setChecked(self.gui_state.send_user_pings)
        self._send_pings_checkbox.stateChanged.connect(self._on_send_pings_checkbox_changed)
        hbox.addWidget(self._send_pings_checkbox)

        hbox.addStretch(1)

    @property
    def consent_to_send_usage_information(self):
        return self._send_pings_checkbox.isChecked()

    def _on_send_pings_checkbox_changed(self):
        self.gui_state.send_user_pings = self._send_pings_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

    def _welcome_to_freemocap_title(self):
        logger.debug("Creating `welcome to freemocap` layout")

        session_title_label = QLabel("Welcome  to  FreeMoCap!")
        session_title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        session_title_label.setStyleSheet("font-size: 54px;")

        return session_title_label

    def _add_freemocap_logo(self):
        freemocap_logo_label = LogoSvgWidget(image_path=PATH_TO_FREEMOCAP_LOGO_SVG)
        self._layout.addStretch(1)
        self._layout.addWidget(freemocap_logo_label)
        self._layout.addStretch(1)



from pathlib import Path
from typing import Union

import numpy as np
import pytest

from freemocap.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)


@pytest.mark.usefixtures("total_body_center_of_mass_data")
def test_total_body_center_of_mass_data_exists(
        total_body_center_of_mass_data: Union[str, Path],
):
    assert Path(
        total_body_center_of_mass_data
    ).is_file(), f"No file found at {total_body_center_of_mass_data}"


@pytest.mark.usefixtures("synchronized_video_folder_path", "total_body_center_of_mass_data")
def test_total_body_center_of_mass_data_shape(
        synchronized_video_folder_path: Union[str, Path],
        total_body_center_of_mass_data: Union[str, Path, np.ndarray],
):
    if isinstance(total_body_center_of_mass_data, (str, Path)):
        total_body_center_of_mass_fr_xyz = np.load(total_body_center_of_mass_data)
    else:
        total_body_center_of_mass_fr_xyz = total_body_center_of_mass_data

    frame_counts = list(get_number_of_frames_of_videos_in_a_folder(synchronized_video_folder_path).values())
    assert (
            len(set(frame_counts)) == 1
    ), f"Videos in {synchronized_video_folder_path} have different frame counts: {frame_counts}"

    number_of_frames = frame_counts[0]

    assert (
            total_body_center_of_mass_fr_xyz.shape[0] == number_of_frames
    ), f"Number of frames in {total_body_center_of_mass_fr_xyz.shape} does not match number of frames of videos in {synchronized_video_folder_path}"

    assert (
            total_body_center_of_mass_fr_xyz.shape[1] == 3
    ), f"total body center of mass should have 3 dimensions (X,Y,Z) but has {total_body_center_of_mass_fr_xyz.shape[1]}"



import rerun as rr
import numpy as np

rr.init("rerun_example_my_data", spawn=True)

positions = np.zeros((10, 3))
positions[:,0] = np.linspace(-10,10,10)

colors = np.zeros((10,3), dtype=np.uint8)
colors[:,0] = np.linspace(0,255,10)

rr.log(
    "my_points",
    rr.Points3D(positions, colors=colors, radii=0.5)
)



import logging
import time

import cv2
import numpy as np
from fastapi import APIRouter, WebSocket
from src.cameras.capture.dataclasses.frame_payload import FramePayload

logger = logging.getLogger(__name__)

cam_ws_router = APIRouter()


async def websocket_send(web_socket: WebSocket, input_payload: FramePayload):
    if not input_payload.success:
        return
    success, frame = cv2.imencode(".png", input_payload.image)
    if not success:
        return
    await web_socket.send_bytes(frame.tobytes())


@cam_ws_router.websocket("/ws/hello_world")
async def preview_webcam(web_socket: WebSocket):
    await web_socket.accept()
    while True:
        last_read = time.perf_counter()
        byte_data = await web_socket.receive_bytes()
        time_since = time.perf_counter()
        print(f"Per Frame Time: {time_since - last_read:.4f}")
        buffer = np.frombuffer(byte_data, dtype="uint8").reshape((600, 800, 4))
        buffer = cv2.cvtColor(buffer, cv2.COLOR_BGR2RGB)
        print(f"image shape: {buffer.shape}")



import logging
import shutil
from pathlib import Path
from typing import Union

logger = logging.getLogger(__name__)


def copy_directory_if_contains_timestamps(source_dir: Union[Path, str], destination_dir: Union[Path, str]) -> bool:
    source_path = Path(source_dir)
    destination_path = Path(destination_dir)

    # Check if the source directory exists
    if not source_path.is_dir():
        raise ValueError(f"{source_dir} is not a valid directory.")

    # Check if the directory contains a directory or file called 'timestamps'
    timestamps_path = source_path / "timestamps"
    if not timestamps_path.exists():
        logger.info(f"{source_dir} does not contain a 'timestamps' directory or file.")
        return False

    # Create the destination directory if it doesn't exist
    destination_path.mkdir(parents=True, exist_ok=True)

    if destination_path.stem == "timestamps":
        destination_path = destination_path.parent

    # Copy the 'timestamps' directory/file to the destination directory
    shutil.copytree(timestamps_path, destination_path / "timestamps", dirs_exist_ok=True)
    return True



import logging
from pathlib import Path

import freemocap

logger = logging.getLogger(__name__)

# directory names
BASE_FREEMOCAP_DATA_FOLDER_NAME = "freemocap_data"
RECORDING_SESSIONS_FOLDER_NAME = "recording_sessions"
CALIBRATIONS_FOLDER_NAME = "calibrations"
LOGS_INFO_AND_SETTINGS_FOLDER_NAME = "logs_info_and_settings"
LOG_FILE_FOLDER_NAME = "logs"
OUTPUT_DATA_FOLDER_NAME = "output_data"
SYNCHRONIZED_VIDEOS_FOLDER_NAME = "synchronized_videos"
ANNOTATED_VIDEOS_FOLDER_NAME = "annotated_videos"
RAW_DATA_FOLDER_NAME = "raw_data"
CENTER_OF_MASS_FOLDER_NAME = "center_of_mass"

# file names
MOST_RECENT_RECORDING_TOML_FILENAME = "most_recent_recording.toml"
LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME = "last_successful_calibration.toml"
GUI_STATE_JSON_FILENAME = "gui_state.json"
DATA_2D_NPY_FILE_NAME = "2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy"

RAW_3D_NPY_FILE_NAME = "3dData_numFrames_numTrackedPoints_spatialXYZ.npy"
REPROJECTION_ERROR_NPY_FILE_NAME = "3dData_numFrames_numTrackedPoints_reprojectionError.npy"
FULL_REPROJECTION_ERROR_NPY_FILE_NAME = "3dData_numCams_numFrames_numTrackedPoints_reprojectionError.npy"
REPROJECTION_FILTERED_PREFIX = "reprojection_filtered_"

DATA_3D_NPY_FILE_NAME = "skeleton_3d.npy"
RIGID_BONES_NPY_FILE_NAME = "rigid_bones_3d.npy"
BODY_3D_DATAFRAME_CSV_FILE_NAME = "body_3d_xyz.csv"
RIGHT_HAND_3D_DATAFRAME_CSV_FILE_NAME = "right_hand_3d_xyz.csv"
LEFT_HAND_3D_DATAFRAME_CSV_FILE_NAME = "left_hand_3d_xyz.csv"
FACE_3D_DATAFRAME_CSV_FILE_NAME = "face_3d_xyz.csv"

TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME = "total_body_center_of_mass_xyz.npy"
SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME = "segmentCOM_frame_joint_xyz.npy"

OLD_DATA_2D_NPY_FILE_NAME = "mediapipe2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy"
OLD_RAW_3D_NPY_FILE_NAME = "mediapipe3dData_numFrames_numTrackedPoints_spatialXYZ.npy"
OLD_REPROJECTION_ERROR_NPY_FILE_NAME = "mediapipe3dData_numFrames_numTrackedPoints_reprojectionError.npy"
OLD_DATA_3D_NPY_FILE_NAME = "mediaPipeSkel_3d_body_hands_face.npy"
OLD_TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME = "total_body_center_of_mass_xyz.npy"
OLD_SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME = "segmentCOM_frame_joint_xyz.npy"

RECORDING_PARAMETERS_JSON_FILE_NAME = "recording_parameters.json"

# Figshare info
FIGSHARE_SAMPLE_ZIP_FILE_URL = "https://figshare.com/ndownloader/files/53681960"
FIGSHARE_TEST_ZIP_FILE_URL = "https://figshare.com/ndownloader/files/53682050"
FREEMOCAP_TEST_DATA_RECORDING_NAME = "freemocap_test_data"
FREEMOCAP_SAMPLE_DATA_RECORDING_NAME = "freemocap_sample_data"

# documentation links
DOCUMENTATION_HOME = "https://freemocap.github.io/documentation/index_md.html"
DOCUMENTATION_WELCOME_URL = "https://freemocap.github.io/documentation/getting-started-index.html"
DOCUMENTATION_SINGLE_CAMERA_URL = "https://freemocap.github.io/documentation/single-camera-recording.html"
DOCUMENTATION_MULTI_CAMERA_URL = "https://freemocap.github.io/documentation/multi-camera-calibration.html"
DOCUMENTATION_PRIVACY_POLICY_URL = "https://freemocap.github.io/documentation/privacy-policy.html"

# logo
PATH_TO_FREEMOCAP_LOGO_SVG = str(Path(freemocap.__file__).parent / "assets/logo/freemocap-logo-black-border.svg")

# progress bars
LOG_VIEW_PROGRESS_BAR_STRING = "Log Progress"

# emoji strings
SPARKLES_EMOJI_STRING = "\U00002728"
SKULL_EMOJI_STRING = "\U0001F480"
DIZZY_EMOJI_STRING = "\U0001F4AB"
EYES_EMOJI_STRING = "\U0001F440"
CAMERA_WITH_FLASH_EMOJI_STRING = "\U0001F4F8"
HAMMER_AND_WRENCH_EMOJI_STRING = "\U0001F6E0"
ROBOT_EMOJI_STRING = "\U0001F916"
THREE_HEARTS_EMOJI_STRING = "\U0001F495"
WIND_EMOJI_STRING = "\U0001F32C"
GEAR_EMOJI_STRING = "\U00002699"
FOLDER_EMOJI_STRING = "\U0001F4C1"
DIRECTORY_EMOJI_STRING = "\U0001F4C1"
COOL_EMOJI_STRING = "\U0001F60E"



from pyqtgraph.parametertree import Parameter
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
    MediapipeTrackingParams,
)

from freemocap.data_layer.recording_models.post_processing_parameter_models import (
    ProcessingParameterModel,
    AniposeTriangulate3DParametersModel,
    PostProcessingParametersModel,
    ButterworthFilterParametersModel,
)

BUTTERWORTH_ORDER = "Order"

BUTTERWORTH_CUTOFF_FREQUENCY = "Cutoff Frequency"

POST_PROCESSING_FRAME_RATE = "Framerate"

BUTTERWORTH_FILTER_TREE_NAME = "Butterworth Filter"

USE_RANSAC_METHOD = "Use RANSAC Method"

ANIPOSE_CONFIDENCE_CUTOFF = "Confidence Threshold Cut-off"

REPROJECTION_ERROR_FILTERING_TREE_NAME = "Reprojection Error Filtering"

RUN_REPROJECTION_ERROR_FILTERING = "Run Reprojection Error Filtering"

REPROJECTION_ERROR_FILTER_THRESHOLD = "Reprojection Error Filter Threshold (%)"

MINIMUM_CAMERAS_TO_REPROJECT = "Minimum Cameras to Reproject"

FLATTEN_SINGLE_CAMERA_DATA = "Flatten Single Camera Data (Recommended)"

ANIPOSE_TREE_NAME = "Anipose Triangulation"

YOLO_CROP_TREE_NAME = "YOLO Crop"

USE_YOLO_CROP_METHOD = "Use YOLO Crop Method"

YOLO_MODEL_SIZE = "YOLO Model Size"

BOUNDING_BOX_BUFFER_METHOD = "Buffer Bounding Box:"

BOUNDING_BOX_BUFFER_PERCENTAGE = "Bounding Box Buffer Percentage"

STATIC_IMAGE_MODE = "Static Image Mode"

MINIUMUM_TRACKING_CONFIDENCE = "Minimum Tracking Confidence"

MINIMUM_DETECTION_CONFIDENCE = "Minimum Detection Confidence"

MEDIAPIPE_MODEL_COMPLEXITY = "Model Complexity"

MEDIAPIPE_TREE_NAME = "Mediapipe"

RUN_IMAGE_TRACKING_NAME = "Run 2d image tracking?"

RUN_3D_TRIANGULATION_NAME = "Run 3d triangulation?"

RUN_BUTTERWORTH_FILTER_NAME = "Run butterworth filter?"

NUMBER_OF_PROCESSES_PARAMETER_NAME = "Max Number of Processes to Use"


# TODO: figure out how to generalize this
def create_mediapipe_parameter_group(
        parameter_model: MediapipeTrackingParams,
) -> Parameter:
    mediapipe_model_complexity_list = [
        "0 (Fastest/Least accurate)",
        "1 (Middle ground)",
        "2 (Slowest/Most accurate)",
    ]
    return Parameter.create(
        name=MEDIAPIPE_TREE_NAME,
        type="group",
        children=[
            dict(
                name=YOLO_CROP_TREE_NAME,
                type="group",
                children=[
                    dict(
                        name=USE_YOLO_CROP_METHOD,
                        type="bool",
                        value=parameter_model.use_yolo_crop_method,
                        tip="If true, `skellytracker` will use YOLO to pre-crop the person from the image before running the `mediapipe` tracker",
                    ),
                    dict(
                        name=YOLO_MODEL_SIZE,
                        type="list",
                        limits=["nano", "small", "medium", "large", "extra_large", "high_res"],
                        value=parameter_model.yolo_model_size,
                        tip="Smaller models are faster but may be less accurate",
                    ),
                    dict(
                        name=BOUNDING_BOX_BUFFER_METHOD,
                        type="list",
                        limits=["By box size", "By image size"],
                        value=parameter_model.buffer_size_method,
                        tip="Buffer bounding box by percentage of either box size or image size",
                    ),
                    dict(
                        name=BOUNDING_BOX_BUFFER_PERCENTAGE,
                        type="int",
                        value=parameter_model.bounding_box_buffer_percentage,
                        limits=(0, 100),
                        step=1,
                        tip="Percentage to increase size of bounding box",
                    ),
                ],
            ),
            dict(
                name=MEDIAPIPE_MODEL_COMPLEXITY,
                type="list",
                limits=mediapipe_model_complexity_list,
                value=mediapipe_model_complexity_list[parameter_model.mediapipe_model_complexity],
                tip="Which Mediapipe model to use - higher complexity is slower but more accurate. "
                    "Variable name in `mediapipe` code: `mediapipe_model_complexity`",
            ),
            dict(
                name=MINIMUM_DETECTION_CONFIDENCE,
                type="float",
                value=parameter_model.min_detection_confidence,
                step=0.05,
                limits=(0.0, 1.0),
                tip="Minimum confidence for a skeleton detection to be considered valid. "
                    "Variable name in `mediapipe` code: `min_detection_confidence`."
                    "NOTE - Never trust a machine learning model's estimates of their own confidence!",
            ),
            dict(
                name=MINIUMUM_TRACKING_CONFIDENCE,
                type="float",
                value=parameter_model.min_tracking_confidence,
                step=0.05,
                limits=(0.0, 1.0),
                tip="Minimum confidence needed to use the previous frame's skeleton estiamte to predict the next one"
                    "Variable name in `mediapipe` code: `min_tracking_confidence`.",
            ),
            dict(
                name=STATIC_IMAGE_MODE,
                type="bool",
                value=parameter_model.static_image_mode,
                tip="If true, the model will process each image independently, without tracking across frames."
                    "I think this is equivalent to setting `min_tracking_confidence` to 0.0"
                    "Variable name in `mediapipe` code: `static_image_mode`",
            ),
        ],
    )


def create_3d_triangulation_parameter_group(
        parameter_model: AniposeTriangulate3DParametersModel = None,
) -> Parameter:
    if parameter_model is None:
        parameter_model = AniposeTriangulate3DParametersModel()

    return Parameter.create(
        name=ANIPOSE_TREE_NAME,
        type="group",
        children=[
            dict(
                name=USE_RANSAC_METHOD,
                type="bool",
                value=parameter_model.use_triangulate_ransac_method,
                tip="If true, use `anipose`'s `triangulate_ransac` method instead of the default `triangulate_simple` method. "
                    "NOTE - Much slower than the 'simple' method, but might be more accurate and better at rejecting bad camera views. Needs more testing and evaluation to see if it's worth it. ",
            ),
            dict(
                name=FLATTEN_SINGLE_CAMERA_DATA,
                type="bool",
                value=parameter_model.flatten_single_camera_data,
                tip="If true, flatten the data from single camera recordings.",
            ),
            dict(
                name=REPROJECTION_ERROR_FILTERING_TREE_NAME,
                type="group",
                children=[
                    dict(
                        name=RUN_REPROJECTION_ERROR_FILTERING,
                        type="bool",
                        value=parameter_model.run_reprojection_error_filtering,
                        tip="If true, run filtering of reprojection error.",
                    ),
                    dict(
                        name=REPROJECTION_ERROR_FILTER_THRESHOLD,
                        type="float",
                        value=parameter_model.reprojection_error_confidence_cutoff,
                        tip="The maximum reprojection error allowed in the data.",
                    ),
                    dict(
                        name=MINIMUM_CAMERAS_TO_REPROJECT,
                        type="int",
                        value=parameter_model.minimum_cameras_to_reproject,
                        tip="The minimum number of cameras to reproject during retriangulation.",
                    ),
                ],
            ),
        ],
    )


def create_post_processing_parameter_group(
        parameter_model: PostProcessingParametersModel = None,
) -> Parameter:
    if parameter_model is None:
        parameter_model = PostProcessingParametersModel()

    return Parameter.create(
        name=BUTTERWORTH_FILTER_TREE_NAME,
        type="group",
        children=[
            dict(
                name=POST_PROCESSING_FRAME_RATE,
                type="float",
                value=parameter_model.butterworth_filter_parameters.sampling_rate,
                tip="Framerate of the recording " "TODO - Calculate this from the recorded timestamps....",
            ),
            dict(
                name=BUTTERWORTH_CUTOFF_FREQUENCY,
                type="float",
                value=parameter_model.butterworth_filter_parameters.cutoff_frequency,
                tip="Oscillations above this frequency will be filtered from the data. ",
            ),
            dict(
                name=BUTTERWORTH_ORDER,
                type="int",
                value=parameter_model.butterworth_filter_parameters.order,
                tip="Order of the filter."
                    "NOTE - I'm not really sure what this parameter does, but this is what I see in other people's Methods sections so....   lol",
            ),
        ],
        tip="Low-pass, zero-lag, Butterworth filter to remove high frequency oscillations/noise from the data. ",
    )


def extract_parameter_model_from_parameter_tree(
        parameter_object: Parameter,
) -> ProcessingParameterModel:
    parameter_values_dictionary = extract_processing_parameter_model_from_tree(parameter_object=parameter_object)

    return ProcessingParameterModel(
        tracking_parameters_model=MediapipeTrackingParams(
            mediapipe_model_complexity=get_integer_from_mediapipe_model_complexity(
                parameter_values_dictionary[MEDIAPIPE_MODEL_COMPLEXITY]
            ),
            min_detection_confidence=parameter_values_dictionary[MINIMUM_DETECTION_CONFIDENCE],
            min_tracking_confidence=parameter_values_dictionary[MINIUMUM_TRACKING_CONFIDENCE],
            static_image_mode=parameter_values_dictionary[STATIC_IMAGE_MODE],
            run_image_tracking=parameter_values_dictionary[RUN_IMAGE_TRACKING_NAME],
            num_processes=parameter_values_dictionary[NUMBER_OF_PROCESSES_PARAMETER_NAME],
            use_yolo_crop_method=parameter_values_dictionary[USE_YOLO_CROP_METHOD],
            yolo_model_size=parameter_values_dictionary[YOLO_MODEL_SIZE],
            buffer_size_method=get_bounding_box_buffer_method_from_string(
                parameter_values_dictionary[BOUNDING_BOX_BUFFER_METHOD]
            ),
            bounding_box_buffer_percentage=parameter_values_dictionary[BOUNDING_BOX_BUFFER_PERCENTAGE],
        ),
        anipose_triangulate_3d_parameters_model=AniposeTriangulate3DParametersModel(
            run_reprojection_error_filtering=parameter_values_dictionary[RUN_REPROJECTION_ERROR_FILTERING],
            reprojection_error_confidence_cutoff=parameter_values_dictionary[REPROJECTION_ERROR_FILTER_THRESHOLD],
            minimum_cameras_to_reproject=parameter_values_dictionary[MINIMUM_CAMERAS_TO_REPROJECT],
            use_triangulate_ransac_method=parameter_values_dictionary[USE_RANSAC_METHOD],
            flatten_single_camera_data=parameter_values_dictionary[FLATTEN_SINGLE_CAMERA_DATA],
            run_3d_triangulation=parameter_values_dictionary[RUN_3D_TRIANGULATION_NAME],
        ),
        post_processing_parameters_model=PostProcessingParametersModel(
            framerate=parameter_values_dictionary[POST_PROCESSING_FRAME_RATE],
            butterworth_filter_parameters=ButterworthFilterParametersModel(
                sampling_rate=parameter_values_dictionary[POST_PROCESSING_FRAME_RATE],
                cutoff_frequency=parameter_values_dictionary[BUTTERWORTH_CUTOFF_FREQUENCY],
                order=parameter_values_dictionary[BUTTERWORTH_ORDER],
            ),
            run_butterworth_filter=parameter_values_dictionary[RUN_BUTTERWORTH_FILTER_NAME],
        ),
    )


def get_integer_from_mediapipe_model_complexity(mediapipe_model_complexity_value: str):
    mediapipe_model_complexity_dictionary = {
        "0 (Fastest/Least accurate)": 0,
        "1 (Middle ground)": 1,
        "2 (Slowest/Most accurate)": 2,
    }
    return mediapipe_model_complexity_dictionary[mediapipe_model_complexity_value]


def get_bounding_box_buffer_method_from_string(buffer_method_string: str) -> str:
    bounding_box_buffer_method_dict = {
        "By box size": "buffer_by_box_size",
        "By image size": "buffer_by_image_size",
    }
    return bounding_box_buffer_method_dict[buffer_method_string]


def extract_processing_parameter_model_from_tree(parameter_object, value_dictionary: dict = None):
    if value_dictionary is None:
        value_dictionary = {}

    for child in parameter_object.children():
        if child.hasChildren():
            extract_processing_parameter_model_from_tree(child, value_dictionary)
        else:
            value_dictionary[child.name()] = child.value()
    return value_dictionary



import logging
import sys
from datetime import datetime
from enum import Enum
from logging.config import dictConfig

from freemocap.system.paths_and_filenames.path_getters import get_log_file_path

# Suppress some annoying log messages
logging.getLogger("tzlocal").setLevel(logging.WARNING)
logging.getLogger("matplotlib").setLevel(logging.WARNING)

log_view_logging_format_string = "[%(asctime)s.%(msecs)04d] [%(levelname)4s] [%(name)s] [Process:%(process)d Thread:%(thread)d] ::::: [%(name)s:%(funcName)s():%(lineno)s] |||| %(message)s."


class LogLevel(Enum):
    TRACE = 5
    DEBUG = logging.DEBUG  # 10
    INFO = logging.INFO  # 20
    SUCCESS = 25
    WARNING = logging.WARNING  # 30
    ERROR = logging.ERROR  # 40
    CRITICAL = logging.CRITICAL  # 50


logging.addLevelName(LogLevel.TRACE.value, "TRACE")
logging.addLevelName(LogLevel.SUCCESS.value, "SUCCESS")


class DeltaTimeFilter(logging.Filter):
    def __init__(self):
        super().__init__()
        self.prev_time = datetime.now().timestamp()

    def filter(self, record):
        current_time = datetime.now().timestamp()
        delta = current_time - self.prev_time
        record.delta_t = f"Δt:{delta:.6f}s"
        self.prev_time = current_time
        return True


class CustomFormatter(logging.Formatter):
    """A custom Formatter class to include microseconds in log timestamps."""

    def formatTime(self, record, datefmt=None):
        created = record.created
        if isinstance(created, float) or isinstance(created, int):
            timestamp = created
        else:
            raise TypeError("Invalid type for 'created'")

        date_format_with_microseconds = "%Y-%m-%dT%H:%M:%S.%f"  # Including microseconds with %f
        return datetime.strftime(datetime.fromtimestamp(timestamp), date_format_with_microseconds)


class LoggerBuilder:
    DEFAULT_LOGGING = {"version": 1, "disable_existing_loggers": False}

    format_string = (
        "[%(asctime)s] [%(delta_t)s] [%(levelname)8s] [%(name)s] "
        "[%(module)s:%(funcName)s():%(lineno)s] "
        "[PID:%(process)d:%(processName)s TID:%(thread)d:%(threadName)s ] %(message)s"
    )

    def __init__(self, level: LogLevel):
        self.default_logging_formatter = CustomFormatter(fmt=self.format_string, datefmt="%Y-%m-%dT%H:%M:%S")
        dictConfig(self.DEFAULT_LOGGING)

        self._set_logging_level(level)

    def _set_logging_level(self, level: LogLevel):
        logging.root.setLevel(level.value)

    def build_file_handler(self):
        file_handler = logging.FileHandler(get_log_file_path(), encoding="utf-8")
        file_handler.setLevel(LogLevel.TRACE.value)
        file_handler.setFormatter(self.default_logging_formatter)
        file_handler.addFilter(DeltaTimeFilter())
        return file_handler

    class ColoredConsoleHandler(logging.StreamHandler):
        COLORS = {
            "TRACE": "\033[37m",  # Dark White (grey)
            "DEBUG": "\033[34m",  # Blue
            "INFO": "\033[96m",  # Cyan
            "SUCCESS": "\033[95m",  # Magenta
            "WARNING": "\033[33m",  # Yellow
            "ERROR": "\033[101m",  # Background Dark Red
        }

        def emit(self, record):
            color_code = self.COLORS.get(record.levelname, "\033[0m")
            formatted_record = color_code + self.format(record) + "\033[0m"

            pid_color = get_hashed_color(record.process)
            tid_color = get_hashed_color(record.thread)

            formatted_record = formatted_record.replace(
                f"PID:{record.process}:{record.processName}",
                pid_color + f"PID:{record.process}:{record.processName}" + "\033[0m",
            )
            formatted_record = formatted_record.replace(
                f"TID:{record.thread}:{record.threadName}",
                tid_color + f"TID:{record.thread}:{record.threadName}" + "\033[0m",
            )

            print(formatted_record)

    def build_console_handler(self):
        console_handler = self.ColoredConsoleHandler(stream=sys.stdout)
        console_handler.setLevel(LogLevel.TRACE.value)
        console_handler.setFormatter(self.default_logging_formatter)
        console_handler.addFilter(DeltaTimeFilter())
        return console_handler

    def configure(self):
        if len(logging.getLogger().handlers) == 0:
            handlers = [self.build_file_handler(), self.build_console_handler()]
            for handler in handlers:
                if handler not in logging.getLogger("").handlers:
                    logging.getLogger("").handlers.append(handler)
        else:
            from freemocap import logger

            logger.info("Logging already configured")


def ensure_min_brightness(value, threshold=50):
    """Ensure the RGB value is above a certain threshold."""
    return max(value, threshold)


def ensure_not_grey(r, g, b, threshold_diff=100):
    """Ensure that the color isn't desaturated grey by making one color component dominant."""
    max_val = max(r, g, b)
    if abs(r - g) < threshold_diff and abs(r - b) < threshold_diff and abs(g - b) < threshold_diff:
        if max_val == r:
            r = 255
        elif max_val == g:
            g = 255
        else:
            b = 255
    return r, g, b


def get_hashed_color(value):
    """Generate a consistent random color for the given value."""
    # Use modulo to ensure it's within the range of normal terminal colors.
    hashed = hash(value) % 0xFFFFFF  # Keep within RGB 24-bit color
    red = ensure_min_brightness(hashed >> 16 & 255)
    green = ensure_min_brightness(hashed >> 8 & 255)
    blue = ensure_min_brightness(hashed & 255)

    red, green, blue = ensure_not_grey(red, green, blue)

    return "\033[38;2;{};{};{}m".format(red, green, blue)


def configure_logging(level: LogLevel = LogLevel.INFO):
    def trace(self, message, *args, **kws):
        if self.isEnabledFor(LogLevel.TRACE.value):
            self._log(LogLevel.TRACE.value, message, args, **kws, stacklevel=2)

    logging.Logger.trace = trace

    def success(self, message, *args, **kws):
        if self.isEnabledFor(LogLevel.SUCCESS.value):
            self._log(LogLevel.SUCCESS.value, message, args, **kws, stacklevel=2)

    logging.Logger.success = success

    builder = LoggerBuilder(level)
    builder.configure()


def log_test_messages(logger):
    logger.trace("This is a TRACE message.")
    logger.debug("This is a DEBUG message.")
    logger.info("This is an INFO message.")
    logger.success("This is a SUCCESS message.")
    logger.warning("This is a WARNING message.")
    logger.error("This is an ERROR message.")
    logger.critical("This is a CRITICAL message.")

    print("----------This is a print message.------------------")

    import time

    for iter in range(1, 10):
        wait_time = iter / 10
        print(f"Testing timestamps (round: {iter}:")
        logger.info("Starting 1 sec timer (Δt should probably be near 0, unless you've got other stuff going on)")
        tic = time.perf_counter_ns()
        time.sleep(wait_time)
        toc = time.perf_counter_ns()
        elapsed_time = (toc - tic) / 1e9
        logger.info(f"Done {wait_time} sec timer - elapsed time:{elapsed_time} (Δt should be ~{wait_time}s)")


if __name__ == "__main__":
    from freemocap import logger

    configure_logging(LogLevel.TRACE)  # Setting the root logger level to TRACE
    log_test_messages(logger)
    logger.success("Logging setup and tests completed. Check the console output and the log file.")



import numpy as np

from freemocap.core_processes.capture_volume_calibration.by_camera_reprojection_filtering import (
    _get_camera_frame_marker_lists_to_reproject,
    _set_unincluded_data_to_nans,
)


def test_get_camera_frame_marker_lists_to_reproject():
    # test removing one camera
    reprojError_cam_frame_marker = np.array(
        [[[1, 5, 3], [4, 5, 6], [7, 8, 9]], [[2, 3, 4], [5, 6, 9], [8, 9, 10]], [[3, 6, 5], [6, 7, 8], [9, 10, 11]]]
    )
    frame_marker_list = [(0, 1), (1, 2)]
    num_cameras_to_remove = 1
    expected_output = ([[2], [1]], [0, 1], [1, 2])
    assert (
            _get_camera_frame_marker_lists_to_reproject(
                reprojError_cam_frame_marker, frame_marker_list, num_cameras_to_remove
            )
            == expected_output
    )
    # test removing two cameras
    frame_marker_list = [(0, 1), (1, 2)]
    num_cameras_to_remove = 2
    expected_output = ([[2, 0], [1, 2]], [0, 1], [1, 2])
    assert (
            _get_camera_frame_marker_lists_to_reproject(
                reprojError_cam_frame_marker, frame_marker_list, num_cameras_to_remove
            )
            == expected_output
    )
    # test removing all cameras
    frame_marker_list = [(0, 1), (1, 2)]
    num_cameras_to_remove = 3
    expected_output = ([[2, 0, 1], [1, 2, 0]], [0, 1], [1, 2])
    assert (
            _get_camera_frame_marker_lists_to_reproject(
                reprojError_cam_frame_marker, frame_marker_list, num_cameras_to_remove
            )
            == expected_output
    )
    # test removing more cameras than dimensions
    frame_marker_list = [(0, 1), (1, 2)]
    num_cameras_to_remove = 4
    # should just return all of the camera sorted by max reprojection error, no need to error
    expected_output = ([[2, 0, 1], [1, 2, 0]], [0, 1], [1, 2])
    assert (
            _get_camera_frame_marker_lists_to_reproject(
                reprojError_cam_frame_marker, frame_marker_list, num_cameras_to_remove
            )
            == expected_output
    )
    # test providing no frames or markers
    frame_marker_list = []
    num_cameras_to_remove = 2
    expected_output = ([], [], [])
    assert (
            _get_camera_frame_marker_lists_to_reproject(
                reprojError_cam_frame_marker, frame_marker_list, num_cameras_to_remove
            )
            == expected_output
    )


def test_set_unincluded_data_to_nans():
    # test indices are turned to nans
    data = np.random.rand(5, 5, 5, 2)
    frames_with_reprojection_error = [1, 2]
    markers_with_reprojection_error = [3, 4]
    cameras_to_remove = [[0, 1], [2, 3]]
    result = _set_unincluded_data_to_nans(data, frames_with_reprojection_error, markers_with_reprojection_error,
                                          cameras_to_remove)
    for cameras, frame, marker in zip(cameras_to_remove, frames_with_reprojection_error,
                                      markers_with_reprojection_error):
        for camera in cameras:
            np.testing.assert_array_equal(result[camera, frame, marker, :], np.full((2,), np.nan))

    # test other indices aren't touched
    data = np.random.rand(5, 5, 5, 2)
    frames_with_reprojection_error = [1, 2]
    markers_with_reprojection_error = [3, 4]
    cameras_to_remove = [[0, 1], [2, 3]]
    original_data = data.copy()
    result = _set_unincluded_data_to_nans(data, frames_with_reprojection_error, markers_with_reprojection_error,
                                          cameras_to_remove)
    mask = np.ones(data.shape, dtype=bool)
    for cameras, frame, marker in zip(cameras_to_remove, frames_with_reprojection_error,
                                      markers_with_reprojection_error):
        for camera in cameras:
            mask[camera, frame, marker, :] = False
    np.testing.assert_array_equal(result[mask], original_data[mask])


if __name__ == "__main__":
    test_get_camera_frame_marker_lists_to_reproject()



import csv
import json
import platform
from pathlib import Path

import numpy as np

from freemocap.diagnostics.calibration.calibration_utils import (
    CharucoNeighborStats, get_neighbor_distances, get_neighbor_stats)


def run(path_to_recording: Path):
    path_to_3d_data = path_to_recording / "output_data" / "charuco_3d_xyz.npy"
    charuco_3d_data = np.load(path_to_3d_data)
    csv_save_path = Path("data_current_calibration.csv")

    path_to_json = path_to_recording / "charuco_board_info.json"
    with open(path_to_json, "r", encoding="utf-8") as fh:
        charuco_board_info = json.load(fh)

    charuco_square_size_mm = charuco_board_info["square_size_mm"]
    number_of_squares_height = charuco_board_info["num_squares_height"]
    number_of_squares_width = charuco_board_info["num_squares_width"]

    distances_between_squares = get_neighbor_distances(
        charuco_3d_data=charuco_3d_data,
        number_of_squares_width=number_of_squares_width,
        number_of_squares_height=number_of_squares_height,
    )

    square_stats: CharucoNeighborStats = get_neighbor_stats(
        distances=distances_between_squares, charuco_square_size_mm=charuco_square_size_mm
    )

    raw_os = platform.system()
    os_map = {"Darwin": "macOS", "Windows": "Windows", "Linux": "Linux"}
    os_name = os_map.get(raw_os, raw_os)

    row = {
        "os": os_name,  # Windows / Linux / Darwin
        "version": "current",  # tag for this run
        "mean_distance": square_stats.mean_distance,
        "median_distance": square_stats.median_distance,
        "std_distance": square_stats.std_distance,
        "mean_error": square_stats.mean_error,
    }

    with open(csv_save_path, "w", newline="") as fh:
        writer = csv.DictWriter(fh, fieldnames=row.keys())
        writer.writeheader()
        writer.writerow(row)

    print(f"✅ one-row CSV written → {csv_save_path}")


if __name__ == "__main__":
    # Use environment detection to handle different runners
    import sys

    if sys.platform.startswith("win"):
        path_to_recording = Path(r"C:\Users\runneradmin\freemocap_data\recording_sessions\freemocap_test_data")
    elif sys.platform.startswith("linux"):
        path_to_recording = Path("/home/runner/freemocap_data/recording_sessions/freemocap_test_data")
    elif sys.platform.startswith("darwin"):
        path_to_recording = Path("/Users/runner/freemocap_data/recording_sessions/freemocap_test_data")
    else:
        raise RuntimeError(f"Unsupported OS: {sys.platform}")

    run(path_to_recording=path_to_recording)



from PySide6.QtWidgets import QMainWindow, QMenuBar

from freemocap.gui.qt.actions_and_menus.actions import Actions
from freemocap.utilities.download_sample_data import DATASETS


class MenuBar(QMenuBar):
    def __init__(self, actions: Actions, parent: QMainWindow = None):
        super().__init__(parent=parent)

        self._add_actions_to_file_menu(actions)

    def _add_actions_to_file_menu(self, actions: Actions):
        """
        based mostly on: https://realpython.com/python-menus-toolbars/
        """

        # file menu
        file_menu = self.addMenu("&File")

        file_menu.addAction(actions.create_new_recording_action)
        file_menu.addAction(actions.load_most_recent_recording_action)
        file_menu.addAction(actions.load_existing_recording_action)
        file_menu.addAction(actions.import_videos_action)
        file_menu.addAction(actions.set_data_folder_action)
        file_menu.addAction(actions.reset_to_defaults_action)
        file_menu.addAction(actions.exit_action)

        # Actions Menu
        actions_menu = self.addMenu("&Controller")
        actions_menu.addAction(actions.kill_running_threads_and_processes_action)
        actions_menu.addAction(actions.reboot_gui_action)

        # Data Menu
        data_menu = self.addMenu("&Data")

        for key in DATASETS:
            action_attr = f"download_{key}_data_action"
            if hasattr(actions, action_attr):
                action = getattr(actions, action_attr)
                data_menu.addAction(action)

        #
        # # navigation menu
        # navigation_menu = QMenu("Na&vigation", parent=self)
        # self.addMenu(navigation_menu)
        #
        # navigation_menu.addAction(self._show_camera_control_panel_action)
        # navigation_menu.addAction(self._show_calibrate_capture_volume_panel_action)
        # navigation_menu.addAction(self._show_motion_capture_videos_panel_action)
        #
        # help menu
        help_menu = self.addMenu("&Help")
        help_menu.addAction(actions.show_release_notes_action)
        help_menu.addAction(actions.open_docs_action)
        help_menu.addAction(actions.freemocap_foundation_action)
        #
        # support menu
        support_menu = self.addMenu("&Support the Freemocap Project")

        support_menu.addAction(actions.donate_action)
        # support_menu.addAction(actions.send_usage_statistics_action)
        # support_menu.addAction(actions.user_survey_action)


if __name__ == "__main__":
    import sys
    from PySide6.QtWidgets import QApplication, QLabel

    app = QApplication(sys.argv)
    _main_window = QMainWindow()
    _main_window.setCentralWidget(QLabel("Henlo fren"))
    _menu_bar = MenuBar(parent=_main_window)
    # _menu_bar = QMenuBar(parent=_main_window)
    # _fake_menu = QMenu("&Fake Menu", parent=_menu_bar)
    # _menu_bar.addMenu(_fake_menu)
    # _fake_menu.addAction("Fake Action")
    _main_window.setMenuBar(_menu_bar)
    _main_window.show()
    sys.exit(app.exec())



import pprint
from typing import List, Dict, Tuple
from typing import Optional, Any

from pydantic import BaseModel, Field, model_validator
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
    MediapipeModelInfo,
)


class Point(BaseModel):
    """
    A point in 3D space at a particular time
    """

    x: Optional[float] = Field(None, description="The X-coordinate of the point")
    y: Optional[float] = Field(None, description="The Y-coordinate of the point")
    z: Optional[float] = Field(None, description="The Z-coordinate of the point")


class ReprojectionError(BaseModel):
    """
    Reprojection error of a 3D point at a particular time
    """

    value: Optional[float] = Field(None, description="Reprojection error of the point")


class VirtualMarkerDefinition(BaseModel):
    """
    A virtual marker, defined by combining multiple markers with weights to generate a new marker/point
    """

    marker_names: List[str] = Field(
        default_factory=list, description="The names of the markers that define this virtual marker"
    )
    marker_weights: List[float] = Field(
        default_factory=list, description="The weights of the markers that define this virtual marker, must sum to 1"
    )

    @model_validator(mode="before")
    def check_weights(cls, values):
        marker_weights = values.get("marker_weights")
        if sum(marker_weights) != 1:
            raise ValueError(f"Marker weights must sum to 1, got {marker_weights}")
        return values


class SegmentSchema(BaseModel):
    """
    A schema for a segment of a skeleton, defined by a set of tracked points and connections between them
    """

    point_names: List[str] = Field(
        default_factory=list, description="The names of the tracked points that define this segment"
    )
    virtual_marker_definitions: Optional[Dict[str, VirtualMarkerDefinition]] = Field(
        default_factory=dict, description="The virtual markers that define this segment"
    )
    connections: List[Tuple[int, int]] = Field(
        default_factory=list, description="The connections between the tracked points"
    )
    parent: Optional[str] = Field(None, description="The name of the parent of this segment")


class SkeletonSchema(BaseModel):
    body: SegmentSchema = Field(default_factory=SegmentSchema, description="The tracked points that define the body")
    hands: Dict[str, SegmentSchema] = Field(
        default_factory=dict, description="The tracked points that define the hands: keys - (left, right)"
    )
    face: SegmentSchema = Field(default_factory=SegmentSchema, description="The tracked points that define the face")

    def __init__(self, schema_dict: Dict[str, Dict[str, Any]]):
        super().__init__()
        self.body = SegmentSchema(**schema_dict["body"])
        self.hands = {hand: SegmentSchema(**hand_schema) for hand, hand_schema in schema_dict["hands"].items()}
        self.face = SegmentSchema(**schema_dict["face"])

    def dict(self):
        d = {}
        d["body"] = self.body.model_dump()
        d["hands"] = {hand: hand_schema.model_dump() for hand, hand_schema in self.hands.items()}
        d["face"] = self.face.model_dump()
        return d


class Timestamps(BaseModel):
    # TODO - Add Unix time and ISO format stuff and whatnot
    mean: Optional[float] = Field(None, description="The mean timestamp for this frame")
    by_camera: Dict[str, Any] = Field(
        default_factory=dict, description="Timestamps for each camera on this frame (key is video name)"
    )


class FrameData(BaseModel):
    """
    The data for a single frame
    """

    timestamps: Timestamps = Field(default_factory=Timestamps, description="Timestamp data")
    tracked_points: Dict[str, Point] = Field(default_factory=dict, description="The points being tracked")
    reprojection_error: Dict[str, ReprojectionError] = Field(
        default_factory=dict, description="Reprojection error of the point being tracked"
    )

    @property
    def tracked_point_names(self):
        return list(self.tracked_points.keys())

    @property
    def timestamp(self):
        return self.timestamps.mean

    def to_dict(self):
        return {
            "timestamps": self.timestamps.model_dump(),
            "tracked_points": {name: point.model_dump() for name, point in self.tracked_points.items()},
            "reprojection_error": {name: error.model_dump() for name, error in self.reprojection_error.items()},
        }


class InfoDict(BaseModel):
    """
    A dictionary of information about this recording, such as the measured segement lengths and the segment connections that we can use to interpret the tracked points (i.e./e.g. how to connect the dots of skeleton)
    """

    # segment_lengths: Dict[str, Any] = Field(default_factory=dict, description="The lengths of the segments of the body")
    schemas: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="The segment connections for the tracked points"
    )


if __name__ == "__main__":
    # nested_dict = create_nested_dict(FramePacket)
    # print(json.dumps(nested_dict, indent=4))

    print("=====================================\n\n===============================")

    skeleton_schema = SkeletonSchema(schema_dict=MediapipeModelInfo.skeleton_schema)

    pprint.pp(skeleton_schema.dict())
